# *****************************************************************************
# * Copyright (c) 2011, 2013-2014 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""Memory service provides basic operations to read/write memory on a target.
.. |fill| replace:: :meth:`~MemoryContext.fill`
.. |get| replace:: :meth:`~MemoryContext.get`
.. |getChildren| replace:: :meth:`~MemoryService.getChildren`
.. |getContext| replace:: :meth:`~MemoryService.getContext`
.. |getMessage| replace:: :meth:`~MemoryError.getMessage`
.. |processes| replace:: :mod:`~tcf.services.processes`
.. |set| replace:: :meth:`~MemoryContext.set`
.. |DoneGetChildren| replace:: :class:`DoneGetChildren`
.. |DoneGetContext| replace:: :class:`DoneGetContext`
.. |DoneMemory| replace:: :class:`DoneMemory`
.. |MemoryContext| replace:: :class:`MemoryContext`
.. |MemoryError| replace:: :class:`MemoryError`
.. |MemoryListener| replace:: :class:`MemoryListener`
Memory Properties
-----------------
.. _Tcf-Memory-Properties:
Properties
^^^^^^^^^^
+-------------------+--------------+------------------------------------------+
| Name              | Type         | Description                              |
+===================+==============+==========================================+
| PROP_ACCESS_TYPES | |list|       | The access types allowed for this        |
|                   |              | context. Target system can support       |
|                   |              | multiple different memory access types,  |
|                   |              | like instruction and data access.        |
|                   |              | Different access types can use different |
|                   |              | logic for address translation and memory |
|                   |              | mapping, so they can end up accessing    |
|                   |              | different data bits, even if address is  |
|                   |              | the same. Each distinct access type      |
|                   |              | should be represented by separate memory |
|                   |              | context. A memory context can represent  |
|                   |              | multiple access types if they are        |
|                   |              | equivalent - all access same memory bits.|
|                   |              | Same data bits can be exposed through    |
|                   |              | multiple memory contexts.                |
|                   |              | See `Access types`_.                     |
+-------------------+--------------+------------------------------------------+
| PROP_ADDRESS_SIZE | |int|        | Size of memory address in bytes.         |
+-------------------+--------------+------------------------------------------+
| PROP_BIG_ENDIAN   | |bool|       | **True** if memory is big-endian.        |
+-------------------+--------------+------------------------------------------+
| PROP_END_BOUND    | |int|        | Highest address (inclusive) which is     |
|                   |              | valid for the context.                   |
+-------------------+--------------+------------------------------------------+
| PROP_ID           | |basestring| | ID of the context.                       |
+-------------------+--------------+------------------------------------------+
| PROP_NAME         | |basestring| | Name of the context, can be used for UI  |
|                   |              | purposes.                                |
+-------------------+--------------+------------------------------------------+
| PROP_PARENT_ID    | |basestring| | ID of a parent context.                  |
+-------------------+--------------+------------------------------------------+
| PROP_PROCESS_ID   | |basestring| | Process ID, see |processes| service.     |
+-------------------+--------------+------------------------------------------+
| PROP_START_BOUND  | |int|        | Lowest address (inclusive) which is      |
|                   |              | valid for the context.                   |
+-------------------+--------------+------------------------------------------+
.. _Tcf-Memory-Access-Types:
Access Types
^^^^^^^^^^^^
All access types are of type |basestring|.
+--------------------+--------------------------------------------------------+
| Name               | Description                                            |
+====================+========================================================+
| ACCESS_CACHE       | Context is a cache.                                    |
+--------------------+--------------------------------------------------------+
| ACCESS_DATA        | Context represents data access.                        |
+--------------------+--------------------------------------------------------+
| ACCESS_HYPERVISOR  | Context represents a hypervisor view to memory.        |
+--------------------+--------------------------------------------------------+
| ACCESS_INSTRUCTION | Context represent instructions fetch access.           |
+--------------------+--------------------------------------------------------+
| ACCESS_IO          | Context represents IO peripherals.                     |
+--------------------+--------------------------------------------------------+
| ACCESS_PHYSICAL    | Context uses physical addresses.                       |
+--------------------+--------------------------------------------------------+
| ACCESS_SUPERVISOR  | Context represents a supervisor (e.g. Linux kernel)    |
|                    | view to memory.                                        |
+--------------------+--------------------------------------------------------+
| ACCESS_TLB         | Context is a TLB memory.                               |
+--------------------+--------------------------------------------------------+
| ACCESS_USER        | Context represents a user (e.g. application running    |
|                    | in Linux) view to memory.                              |
+--------------------+--------------------------------------------------------+
| ACCESS_VIRTUAL     | Context uses virtual addresses.                        |
+--------------------+--------------------------------------------------------+
.. _Tcf-Memory-Command-Modes:
Command Modes
^^^^^^^^^^^^^
All command modes are of type |int|. For the |get|, |fill| or |set| methods,
the possible modes are:
+----------------------+------------------------------------------------------+
| Name                 | Description                                          |
+======================+======================================================+
| MODE_CONTINUEONERROR | Carry on when some of the memory cannot be accessed  |
|                      | and return |MemoryError| at the end if any of the    |
|                      | bytes were not processed correctly.                  |
+----------------------+------------------------------------------------------+
| MODE_VERIFY          | Verify result of memory operations (by reading and   |
|                      | comparing).                                          |
+----------------------+------------------------------------------------------+
Service Methods
---------------
.. autodata:: NAME
.. autoclass:: MemoryService
addListener
^^^^^^^^^^^
.. automethod:: MemoryService.addListener
fill
^^^^
.. automethod:: MemoryContext.fill
get
^^^
.. automethod:: MemoryContext.get
getChildren
^^^^^^^^^^^
.. automethod:: MemoryService.getChildren
getContext
^^^^^^^^^^
.. automethod:: MemoryService.getContext
getName
^^^^^^^
.. automethod:: MemoryService.getName
removeListener
^^^^^^^^^^^^^^
.. automethod:: MemoryService.removeListener
set
^^^
.. automethod:: MemoryContext.set
Callback Classes
----------------
DoneGetChildren
^^^^^^^^^^^^^^^
.. autoclass:: DoneGetChildren
    :members:
DoneGetContext
^^^^^^^^^^^^^^
.. autoclass:: DoneGetContext
    :members:
DoneMemory
^^^^^^^^^^
.. autoclass:: DoneMemory
    :members:
Listener
--------
MemoryListener
^^^^^^^^^^^^^^
.. autoclass:: MemoryListener
    :members:
Helper classes
--------------
ErrorOffset
^^^^^^^^^^^
.. autoclass:: ErrorOffset
    :members:
MemoryContext
^^^^^^^^^^^^^
.. autoclass:: MemoryContext
    :members:
MemoryError
^^^^^^^^^^^
.. autoclass:: MemoryError
    :members:
"""

from .. import services

NAME = "Memory"
"""Memory service name."""

# Context property names.

PROP_ID = "ID"
PROP_PARENT_ID = "ParentID"
PROP_PROCESS_ID = "ProcessID"
PROP_BIG_ENDIAN = "BigEndian"
PROP_ADDRESS_SIZE = "AddressSize"
PROP_NAME = "Name"
PROP_START_BOUND = "StartBound"
PROP_END_BOUND = "EndBound"
PROP_ACCESS_TYPES = "AccessTypes"

ACCESS_INSTRUCTION = "instruction"
ACCESS_DATA = "data"
ACCESS_IO = "io"
ACCESS_USER = "user"
ACCESS_SUPERVISOR = "supervisor"
ACCESS_HYPERVISOR = "hypervisor"
ACCESS_VIRTUAL = "virtual"
ACCESS_PHYSICAL = "physical"
ACCESS_CACHE = "cache"
ACCESS_TLB = "tlb"

MODE_CONTINUEONERROR = 0x1
MODE_VERIFY = 0x2
MODE_BYPASS_ADDR_CHECK = 0x4
MODE_BYPASS_CACHE_SYNC = 0x8


class MemoryContext(object):
    """Memory context class.
    :param props: Properties to initialise this memory context with. See
                  `Properties`_.
    :type props: |dict|
    """

    def __init__(self, props):
        self._props = props or {}

    def __str__(self):
        return "[Memory Context %s]" % self._props

    def getProperties(self):
        """Get context properties. See `Properties`_ definitions for property
        names.
        Context properties are read only, clients should not try to modify
        them.
        :returns: A |dict| of context properties.
        """
        return self._props

    def getID(self):
        """Retrieve context ID.
        :returns: A |basestring| representing this memory context ID or
                  **None**.
        """
        return self._props.get(PROP_ID)

    def getParentID(self):
        """Retrieve parent context ID.
        :returns: A |basestring| representing this memory context parent ID or
                  **None**.
        """
        return self._props.get(PROP_PARENT_ID)

    def getProcessID(self):
        """Retrieve context process ID.
        :returns: A |basestring| representing this memory context process ID or
                  **None**.
        """
        return self._props.get(PROP_PROCESS_ID)

    def isBigEndian(self):
        """Get memory endianness.
        :returns: A |bool| stating if memory is big-endian.
        """
        return self._props.get(PROP_BIG_ENDIAN, False)

    def getAddressSize(self):
        """Get memory address size.
        :returns: An |int| representing the number of bytes used to store
                  memory address value.
        """
        return self._props.get(PROP_ADDRESS_SIZE, 0)

    def getName(self):
        """Get memory context name.
        The name can be used for UI purposes.
        :returns: A |basestring| representing this context name or **None**.
        """
        return self._props.get(PROP_NAME)

    def getStartBound(self):
        """Get lowest address (inclusive) which is valid for the context.
        :returns: An |int| representing the lowest address or **None**.
        """
        return self._props.get(PROP_START_BOUND)

    def getEndBound(self):
        """Get highest address (inclusive) which is valid for the context.
        :returns: An |int| representing the highest address or **None**.
        """
        return self._props.get(PROP_END_BOUND)

    def getAccessTypes(self):
        """Get the access types allowed for this context.
        :returns: A |dict| of access type names or **None**. See
                 `Access Types`_.
        """
        return self._props.get(PROP_ACCESS_TYPES)

    def set(self, addr, ws, buf, offs, sz, mode, done):  # @ReservedAssignment
        """Set target memory.
        If *ws* is 0 it means client does not care about word size.
        :param addr: Address to set for this memory context.
        :type addr: |int|
        :param ws: Word size (alignement) to use for setting memory.
        :type ws: |int|
        :param buf: Data to be written in memory.
        :type buf: |bytearray|
        :param offs: Offset to seek in *buf*.
        :type offs: |int|
        :param sz: Number of bytes to write in memory.
        :type sz: |int|
        :param mode: Memory write mode. See `Command Modes`_.
        :type mode: |int|
        :returns: pending command handle.
        """
        raise NotImplementedError("Abstract method")

    def get(self, addr, word_size, buf, offs, size, mode, done):
        """Read target memory.
        :param addr: Address to get from this memory context.
        :type addr: |int|
        :param ws: word size (alignement) to use for getting memory.
        :type ws: |int|
        :param buf: read data buffer.
        :type buf: |bytearray|
        :param offs: Offset to seek in *buf* to write retrieved data.
        :type offs: |int|
        :param sz: Number of bytes to read in memory.
        :type sz: |int|
        :param mode: Memory read mode. See `Command Modes`_.
        :type mode: |int|
        :returns: pending command handle.
        """
        raise NotImplementedError("Abstract method")

    def fill(self, addr, word_size, value, size, mode, done):
        """Fill target memory with given pattern.
        :param addr: Address to get from this memory context.
        :type addr: |int|
        :param word_size: Word size (alignement) to use for filling memory.
        :type word_size: |int|
        :param buf: Data to be fill memory with.
        :type buf: |bytearray|
        :param offs: Offset to seek in *buf*.
        :type offs: |int|
        :param sz: Number of bytes to write in memory.
        :type sz: |int|
        :param mode: Memory write mode. See `Command Modes`_.
        :type mode: |int|
        :returns: pending command handle.
        """
        raise NotImplementedError("Abstract method")


class DoneMemory(object):
    """Client call back interface for |fill|, |get| and |set| commands."""

    def doneMemory(self, token, error):
        """Called when memory operation command command is done.
        :param token: Command handle.
        :param error: Error object or **None**.
        """
        pass


class MemoryError(Exception):  # @ReservedAssignment
    """Memory errors class.
    Empty class implementing the |Exception| class.
    """
    pass


class ErrorOffset(object):
    """Error offset may be implemented by |MemoryError| object, which is
    returned by |get|, |set| and |fill| commands.
    The command |get|, |set| and |fill| return this exception when memory
    operation failed for some but not all bytes, and **MODE_CONTINUEONERROR**
    has been set in *mode*. (For example, when only part of the request
    translates to valid memory addresses.)
    Method |getMessage| can be used for generalized message of the possible
    reasons of partial memory operation.
    """
    # Error may have per byte information
    BYTE_VALID = 0x00
    BYTE_UNKNOWN = 0x01  # e.g. out of range
    BYTE_INVALID = 0x02
    BYTE_CANNOT_READ = 0x04
    BYTE_CANNOT_WRITE = 0x08

    RANGE_KEY_ADDR = "addr"
    RANGE_KEY_SIZE = "size"
    RANGE_KEY_STAT = "stat"
    RANGE_KEY_MSG = "msg"

    def getStatus(self, offset):
        """Get the error status.
        :param offset: offset to get error status for
        :type offset: |int|
        :returns: The error status for the error which happened at given
                  *offset*.
        """
        raise NotImplementedError("Abstract method")

    def getMessage(self, offset):
        """Get the error message.
        :param offset: Offset to get error message for.
        :type offset: |int|
        :returns: The error message for the error which happened at given
                  *offset*.
        """
        raise NotImplementedError("Abstract method")


class MemoryService(services.Service):
    """TCF memory service interface."""

    def getName(self):
        """Get this service name.
        :returns: The value of string :const:`NAME`
        """
        return NAME

    def getContext(self, contextID, done):
        """Retrieve context info for given context ID.
        :param contextID: ID of the memory context to get.
        :type contextID: |basestring|
        :param done: Call back interface called when operation is completed.
        :type done: |DoneGetContext|
        :returns: Pending command handle.
        """
        raise NotImplementedError("Abstract method")

    def getChildren(self, parent_context_id, done):
        """Retrieve contexts available for memory commands.
        A context corresponds to an execution thread, process, address space,
        etc.
        A context can belong to a parent context. Contexts hierarchy can be
        simple plain list or it can form a tree. It is up to target agent
        developers to choose layout that is most descriptive for a given
        target. Context IDs are valid across all services. In other words, all
        services access same hierarchy of contexts, with same IDs, however,
        each service accesses its own subset of context's attributes and
        functionality, which is relevant to that service.
        :param parent_context_id: parent context ID. Can be **None** - to
                                  retrieve top level of the hierarchy, or one
                                  of context IDs retrieved by previous
                                  |getChildren| commands.
        :type parent_context_id: |basestring|
        :param done: call back interface called when operation is completed.
        :type done: |DoneGetContext|
        :returns: Pending command handle.
        """
        raise NotImplementedError("Abstract method")

    def addListener(self, listener):
        """Add memory service event listener.
        :param listener: Event listener implementation.
        :type listener: |MemoryListener|
        """
        raise NotImplementedError("Abstract method")

    def removeListener(self, listener):
        """Remove memory service event listener.
        :param listener: Event listener implementation.
        :type listener: |MemoryListener|
        """
        raise NotImplementedError("Abstract method")


class MemoryListener(object):
    """Memory event listener is notified when memory context hierarchy changes,
    and when memory is modified by memory service commands.
    """

    def contextAdded(self, contexts):
        """Called when a new memory access context(s) is created.
        :param contexts: A list of |MemoryContext| properties which have been
                         added to memory space.
        :type contexts: |list|
        """
        pass

    def contextChanged(self, contexts):
        """Called when a memory access context(s) properties changed.
        :param contexts: A list of |MemoryContext| properties which have been
                         changed in memory space.
        :type contexts: |list|
        """
        pass

    def contextRemoved(self, context_ids):
        """Called when memory access context(s) is removed.
        :param context_ids: A list of the IDs of memory contexts which have
                            been removed from memory space.
        :type context_ids: |list|
        """
        pass

    def memoryChanged(self, context_id, addr, size):
        """Called when target memory content was changed and clients need to
        update themselves.
        Clients, at least, should invalidate corresponding cached memory data.
        .. note:: Not every change is notified - it is not possible, only
                  those, which are not caused by normal execution of the
                  debuggee.
        :param context_id: ID of the changed memory context.
        :type context_id: |basestring|
        :param addr: Address at which memory has been changed. A value of
                     **None** means that this address is unknown.
        :type addr: |int|
        :param size: Size of the changed memory. A value of **None** means that
                     this size is unknown.
        :type size: |int| or **None**
        """
        pass


class DoneGetContext(object):
    """Client call back interface for |getContext|."""

    def doneGetContext(self, token, error, context):
        """Called when memory context retrieval is done.
        :param token: Command handle.
        :param error: Error description if operation failed, **None** if
                      succeeded.
        :param context: The so retrieved memory context description.
        :type context: |MemoryContext|
        """
        pass


class DoneGetChildren(object):
    """Client call back interface for |getChildren|."""

    def doneGetChildren(self, token, error, context_ids):
        """Called when memory IDs retrieval is done.
        :param token: Command handle.
        :param error: Error description if operation failed, **None** if
                      succeeded.
        :param context_ids: A list of available context IDs.
        :type context_ids: |list|
        """
        pass



# *****************************************************************************
# * Copyright (c) 2011, 2013, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

import threading

from .. import protocol, channel


class DelegatingEventListener(channel.EventListener):
    def __init__(self, _callable):
        self._callable = _callable

    def event(self, name, data):
        try:
            args = channel.fromJSONSequence(data)
            self._callable(self.svc_name, name, *args)
        except Exception as x:
            protocol.log("Error decoding event data", x)


def _print_event(service, name, *args):
    print("Event: %s.%s%s" % (service, name, tuple(args)))


def get_event_printer():
    return DelegatingEventListener(_print_event)


class EventRecorder(object):
    def __init__(self, channel):
        self._channel = channel
        self._events = []
        self._listeners = {}
        self._lock = threading.RLock()
        self._filter = None

    def __del__(self):
        if self._channel.state == channel.STATE_OPEN:
            self.stop()

    def record(self, service, enable=True):
        with self._lock:
            listener = self._listeners.get(service)
            if listener:
                if not enable:
                    protocol.invokeLater(self._channel.removeEventListener,
                                         service, listener)
            elif enable:
                recorder = self

                class Listener(channel.EventListener):
                    def event(self, name, data):
                        e = Event(service, name, data)
                        recorder._event(e)
                listener = Listener()
                self._listeners[service] = listener
                protocol.invokeLater(self._channel.addEventListener, service,
                                     listener)
            self._recording = enable

    def stop(self, service=None):
        if service:
            self.record(service, False)
        else:
            for service in list(self._listeners.keys()):
                self.record(service, False)

    def get(self):
        with self._lock:
            events = self._events
            self._events = []
        return events

    def _event(self, e):
        with self._lock:
            self._events.append(e)

    def __str__(self):
        events = self.get()
        return "\n".join(map(str, events))
    __repr__ = __str__


class Event(object):
    def __init__(self, service, name, data):
        self.service = service
        self.name = name
        try:
            self.args = channel.fromJSONSequence(data)
        except Exception as x:
            protocol.log("Error decoding event data", x)

    def __str__(self):
        return "Event: %s.%s%s" % (self.service, self.name, tuple(self.args))
    __repr__ = __str__



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""Both hosts and targets are represented by objects implementing Peer
interface.

A peer can act as host or target depending on services it implements. List of
currently known peers can be retrieved by calling |getPeers|.

A TCF agent houses one or more service managers. A service manager has a one or
more services to expose to the world. The service manager creates one or more
peers to represent itself, one for every access path the agent is reachable by.

For example, in agents accessible via TCP/IP, the service manager would create
a peer for every subnet it wants to participate in. All peers of particular
service manager represent identical sets of services.

.. |getPeers| replace:: :meth:`tcf.services.locator.LocatorService.getPeers`
.. |invokeAndWait| replace:: :func:`~tcf.protocol.invokeAndWait`
.. |AssertionError| replace:: :exc:`~exception.AssertionError`

Properties
----------
+-------------------------+--------------+------------------------------------+
| Name                    | Type         | Description                        |
+=========================+==============+====================================+
| ATTR_ID                 | |basestring| | Peer unique ID.                    |
+-------------------------+--------------+------------------------------------+
| ATTR_SERVICE_MANAGER_ID | |basestring| | Unique ID of service manager that  |
|                         |              | is represented by this peer.       |
+-------------------------+--------------+------------------------------------+
| ATTR_AGENT_ID           | |basestring| | Agent unique ID.                   |
+-------------------------+--------------+------------------------------------+
| ATTR_NAME               | |basestring| | Peer name.                         |
+-------------------------+--------------+------------------------------------+
| ATTR_OS_NAME            | |basestring| | Name of the peer operating system. |
+-------------------------+--------------+------------------------------------+
| ATTR_TRANSPORT_NAME     | |basestring| | Transport name, for example TCP,   |
|                         |              | SSL ...                            |
+-------------------------+--------------+------------------------------------+
| ATTR_PROXY              | |bool|       | If present, indicates that the peer|
|                         |              | can forward traffic to other peers.|
+-------------------------+--------------+------------------------------------+
| ATTR_IP_HOST            | |basestring| | Host DNS name or IP address.       |
+-------------------------+--------------+------------------------------------+
| ATTR_IP_ALIASES         | |list|       | Optional list of host aliases.     |
+-------------------------+--------------+------------------------------------+
| ATTR_IP_ADDRESSES       | |list|       | Optional list of host addresses.   |
+-------------------------+--------------+------------------------------------+
| ATTR_IP_PORT            | |int|        | IP port number, must be decimal    |
|                         |              | number.                            |
+-------------------------+--------------+------------------------------------+


Classes
-------
Peer
^^^^
.. autoclass:: Peer
    :members:
"""

import os
import time
from . import protocol, transport, services, channel
from .services import locator

ATTR_ID = "ID"
ATTR_SERVICE_MANAGER_ID = "ServiceManagerID"
ATTR_AGENT_ID = "AgentID"
ATTR_NAME = "Name"
ATTR_OS_NAME = "OSName"
ATTR_TRANSPORT_NAME = "TransportName"
ATTR_PROXY = "Proxy"
ATTR_IP_HOST = "Host"
ATTR_IP_ALIASES = "Aliases"
ATTR_IP_ADDRESSES = "Addresses"
ATTR_IP_PORT = "Port"


class Peer(object):
    """TCF agent peer.

    :param attrs: Peer attributes to initialise this peer with.
    :type attrs: |dict|
    """
    def __init__(self, attrs):
        self.attrs = attrs or {}

    def __repr__(self):
        return str(self.__class__.__name__) + '(' + repr(self.attrs) + ')'

    def __str__(self):
        res = str(self.__class__.__name__) + ' ['
        res += ATTR_ID + '=' + str(self.attrs.get(ATTR_ID))
        res += ', ' + ATTR_NAME + '=' + str(self.attrs.get(ATTR_NAME))
        res += ', ' + ATTR_PROXY + '=' + str(self.attrs.get(ATTR_PROXY))
        res += ', ' + ATTR_OS_NAME + '=' + str(self.attrs.get(ATTR_OS_NAME))
        res += ', ' + ATTR_AGENT_ID + '=' + str(self.attrs.get(ATTR_AGENT_ID))
        res += ', ' + ATTR_IP_ADDRESSES + '=' + \
               str(self.attrs.get(ATTR_IP_ADDRESSES))
        res += ', ' + ATTR_IP_ALIASES + '=' + \
               str(self.attrs.get(ATTR_IP_ALIASES))
        res += ', ' + ATTR_IP_HOST + '=' + str(self.attrs.get(ATTR_IP_HOST))
        res += ', ' + ATTR_IP_PORT + '=' + str(self.attrs.get(ATTR_IP_PORT))
        res += ', ' + ATTR_SERVICE_MANAGER_ID + '=' + \
               str(self.attrs.get(ATTR_SERVICE_MANAGER_ID))
        res += ', ' + ATTR_TRANSPORT_NAME + '=' + \
               str(self.attrs.get(ATTR_TRANSPORT_NAME))
        res += ']'
        return res

    def getAddresses(self):
        """Get this peer IP addresses.

        :returns: A |list| representing this peer possible IP addresses, or an
                  empty list.
        """
        return self.attrs.get(ATTR_IP_ADDRESSES, [])

    def getAgentID(self):
        """Get this peer agent ID.

        :returns: This peer's agent unique ID or an empty string

        :raises: An |AssertionError| if this method is not called from the
                 dispatch thread.

        .. seealso:: |invokeAndWait|
        """
        assert protocol.isDispatchThread()
        return self.attrs.get(ATTR_AGENT_ID, '')

    def getAliases(self):
        """Get this peer host name aliases.

        :returns: A |list| representing this peer possible host name aliases,
                  or an empty list.
        """
        return self.attrs.get(ATTR_IP_ALIASES, [])

    def getAttributes(self):
        """Get this peer attributes.

        :returns: A |dict| of this peer attributes.
        """
        return self.attrs

    def getID(self):
        """Get this peer ID.

        :returns: A |basestring| representing this peer unique ID, or an empty
                  string.
        """
        return self.attrs.get(ATTR_ID, '')

    def getHost(self):
        """Get this peer host name.

        :returns: A |basestring| representing this peer host name, or an empty
                  string.
        """
        return self.attrs.get(ATTR_IP_HOST, '')

    def getName(self):
        """Get this peer name.

        :returns: A |basestring| representing this peer name, or an empty
                  string.
        """
        return self.attrs.get(ATTR_NAME, '')

    def getOSName(self):
        """Get this peer's agent OS name.

        :returns: A |basestring| representing this peer agent OS name or an
                  empty string.
        """
        return self.attrs.get(ATTR_OS_NAME, '')

    def getPort(self):
        """Get this peer's communication port.

        :returns: A |basestring| representing this peer communication port
                  number.
        """
        return self.attrs.get(ATTR_IP_PORT, '')

    def getServiceManagerID(self):
        """Get the peer service manager ID.

        :returns: A |basestring| representing this peer's service manager
                  unique ID, or an empty string.

        :raises: An |AssertionError| if this method is not
                 called from the dispatch thread.
        """
        assert protocol.isDispatchThread()
        return self.attrs.get(ATTR_SERVICE_MANAGER_ID, '')

    def getTransportName(self):
        """Get this peer's transport name.

        :returns: this peer agent transport name or an empty string

        .. seealso:: |invokeAndWait|
        """
        return self.attrs.get(ATTR_TRANSPORT_NAME, '')

    def isProxy(self):
        """Check if this peer is a proxy.

        :returns: A |bool| stating if this peer is a proxy or not.
        """
        return self.attrs.get(ATTR_PROXY, False)

    def openChannel(self):
        """Open channel to communicate with this peer.

        .. note:: The channel is not fully open yet when this method returns.
                  Its state is channel.STATE_OPENING.
                  Protocol.ChannelOpenListener and channel.ChannelListener
                  listeners will be called when the channel will change state
                  to open or closed. Clients are supposed to register
                  channel.ChannelListener right after calling openChannel(),
                  or, at least, in same dispatch cycle.

        For example:

        .. code-block:: python

            import tcf.peer as peer

            channel = peer.openChannel()
            channel.addChannelListener(...)
        """
        raise RuntimeError("Abstract method")


class TransientPeer(Peer):
    """
    Transient implementation of IPeer interface.
    Objects of this class are not tracked by Locator service.
    See AbstractPeer for IPeer objects that should go into the Locator table.
    """
    def __init__(self, attrs):
        self.rw_attrs = {}
        self.rw_attrs.update(attrs)
        # TODO readonly map
        ro_attrs = {}
        ro_attrs.update(self.rw_attrs)
        super(TransientPeer, self).__init__(ro_attrs)

    def openChannel(self):
        return transport.openChannel(self)


class LocalPeer(TransientPeer):
    """
    LocalPeer object represents local end-point of TCF communication channel.
    There should be exactly one such object in a TCF agent.
    The object can be used to open a loop-back communication channel that
    allows the agent to access its own services same way as remote services.
    Note that "local" here is relative to the agent, and not same as in
    "local host".
    """
    def __init__(self):
        super(LocalPeer, self).__init__(self.createAttributes())

    def createAttributes(self):
        attrs = {
            ATTR_ID: "TCFLocal",
            ATTR_SERVICE_MANAGER_ID: services.getServiceManagerID(),
            ATTR_AGENT_ID: protocol.getAgentID(),
            ATTR_NAME: "Local Peer",
            ATTR_OS_NAME: os.name,
            ATTR_TRANSPORT_NAME: "Loop"
        }
        return attrs


class AbstractPeer(TransientPeer):
    """
    Abstract implementation of IPeer interface.
    Objects of this class are stored in Locator service peer table.
    The class implements sending notification events to Locator listeners.
    See TransientPeer for IPeer objects that are not stored in the Locator
    table.
    """

    def __init__(self, attrs):
        super(AbstractPeer, self).__init__(attrs)
        assert protocol.isDispatchThread()
        self.last_heart_beat_time = 0
        _id = self.getID()
        assert _id
        peers = protocol.getLocator().getPeers()
        if isinstance(peers.get(_id), RemotePeer):
            peers.get(_id).dispose()
        assert _id not in peers
        peers[_id] = self
        self.sendPeerAddedEvent()

    def dispose(self):
        assert protocol.isDispatchThread()
        _id = self.getID()
        assert _id
        peers = protocol.getLocator().getPeers()
        assert peers.get(_id) == self
        del peers[_id]
        self.sendPeerRemovedEvent()

    def onChannelTerminated(self):
        # A channel to this peer was terminated:
        # not delaying next heart beat helps client to recover much faster.
        self.last_heart_beat_time = 0

    def updateAttributes(self, attrs):
        equ = True
        assert attrs.get(ATTR_ID) == self.rw_attrs.get(ATTR_ID)
        for key in list(self.rw_attrs.keys()):
            if self.rw_attrs.get(key) != attrs.get(key):
                equ = False
                break
        for key in list(attrs.keys()):
            if attrs.get(key) != self.rw_attrs.get(key):
                equ = False
                break
        timeVal = int(time.time() * 1000)
        if not equ:
            self.rw_attrs.clear()
            self.rw_attrs.update(attrs)
            for l in protocol.getLocator().getListeners():
                try:
                    l.peerChanged(self)
                except Exception as x:
                    protocol.log("Unhandled exception in Locator listener", x)
            try:
                args = [self.rw_attrs]
                protocol.sendEvent(locator.NAME, "peerChanged",
                                   channel.toJSONSequence(args))
            except IOError as x:
                protocol.log("Locator: failed to send 'peerChanged' event", x)
            self.last_heart_beat_time = timeVal
        elif self.last_heart_beat_time + \
                int(locator.DATA_RETENTION_PERIOD / 4) < timeVal:
            for l in protocol.getLocator().getListeners():
                try:
                    l.peerHeartBeat(attrs.get(ATTR_ID))
                except Exception as x:
                    protocol.log("Unhandled exception in Locator listener", x)
            try:
                args = [self.rw_attrs.get(ATTR_ID)]
                protocol.sendEvent(locator.NAME, "peerHeartBeat",
                                   channel.toJSONSequence(args))
            except IOError as x:
                protocol.log(
                    "Locator: failed to send 'peerHeartBeat' event", x)
            self.last_heart_beat_time = timeVal

    def sendPeerAddedEvent(self):
        for l in protocol.getLocator().getListeners():
            try:
                l.peerAdded(self)
            except Exception as x:
                protocol.log("Unhandled exception in Locator listener", x)
        try:
            args = [self.rw_attrs]
            protocol.sendEvent(locator.NAME, "peerAdded",
                               channel.toJSONSequence(args))
        except IOError as x:
            protocol.log("Locator: failed to send 'peerAdded' event", x)
        self.last_heart_beat_time = int(time.time() * 1000)

    def sendPeerRemovedEvent(self):
        for l in protocol.getLocator().getListeners():
            try:
                l.peerRemoved(self.rw_attrs.get(ATTR_ID))
            except Exception as x:
                protocol.log("Unhandled exception in Locator listener", x)
        try:
            args = [self.rw_attrs.get(ATTR_ID)]
            protocol.sendEvent(locator.NAME, "peerRemoved",
                               channel.toJSONSequence(args))
        except IOError as x:
            protocol.log("Locator: failed to send 'peerRemoved' event", x)


class RemotePeer(AbstractPeer):
    """
    RemotePeer objects represent TCF agents that Locator service discovered on
    local network.
    This includes both local host agents and remote host agents.
    Note that "remote peer" means any peer accessible over network,
    it does not imply the agent is running on a "remote host".
    If an agent binds multiple network interfaces or multiple ports, it can be
    represented by multiple RemotePeer objects - one per each network
    address/port combination.
    RemotePeer objects life cycle is managed by Locator service.
    """

    def __init__(self, attrs):
        super(RemotePeer, self).__init__(attrs)
        self.last_update_time = int(time.time() * 1000)

    def updateAttributes(self, attrs):
        super(RemotePeer, self).updateAttributes(attrs)
        self.last_update_time = int(time.time() * 1000)

    def getLastUpdateTime(self):
        return self.last_update_time



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import code
from typing import Callable
from threading import Condition
from chipscopy import client
from chipscopy.tcf import protocol, peer
from chipscopy.client.util import config  # noqa: F401

# noinspection PyUnresolvedReferences
from chipscopy.tcf.services.remote import (  # noqa: F401
    XicomEverestProxy,
    svfProxy,
    LocatorProxy,
    MemoryProxy,
)

# noinspection PyUnresolvedReferences
from chipscopy.proxies import DebugCoreProxy, DebugCorePollingProxy  # noqa: F401

try:
    import readline

    readline.parse_and_bind("tab: complete")
except Exception as e:
    pass


def start_interactive(local=None):
    console = code.InteractiveConsole(local)
    console.push("from chipscopy.client.util import *")
    console.push("from chipscopy.client.util.config import *")
    console.push("init_config()")
    console.push("from chipscopy.utils.logger import log")
    console.push("from sys import exit")
    console.interact()


def connect_hw(url):
    from .config import init_config

    init_config()
    return client.connect(url)


def get_connections():
    return client.connections()


def process_param_str(params, default_port="3121"):
    parts = params.split(":", 3)
    if len(parts) < 3:
        parts.insert(0, "TCP")
    if len(parts) < 3:
        parts.append(default_port)
    return ":".join(parts)


def parse_params(paramStr):
    args = paramStr.split(":")
    if len(args) != 3:
        raise ValueError("Expected format: <transport>:<host>:<port>")
    transp, host, port = args
    return {
        peer.ATTR_ID: paramStr,
        peer.ATTR_IP_HOST: host,
        peer.ATTR_IP_PORT: port,
        peer.ATTR_TRANSPORT_NAME: transp,
    }


def sync_call(command: Callable, *args, **kwargs):
    """
    Performs a synchronous function call with a done callback that's intended to run on the TCF dispatch thread.
    :param command: Function to be run on the TCF dispatch thread
    :param args: args of function
    :param kwargs: kwargs of function
    :return: Result of function, or exception if error
    """
    cond = Condition()
    call_error = None
    call_result = None

    def done_call(token, error, result):
        nonlocal call_error
        nonlocal call_result
        call_error = error
        call_result = result
        with cond:
            cond.notify()

    kwargs["done"] = done_call
    with cond:
        protocol.invokeLater(command, *args, **kwargs)
        cond.wait()

    if call_error:
        if isinstance(call_error, str):
            call_error = Exception(call_error)
        raise call_error
    return call_result



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict

from chipscopy import dm
from chipscopy.client.core import CoreClient
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.proxies.SysMonProxy import SYSMON_SERVICE_NAME


class SysMonCoreClient(CoreClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.type == "sysmon"

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(SYSMON_SERVICE_NAME)

    def post_init(self):
        self.refresh_property_group(["status"])

    def refresh_property_group(self, groups: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_property_group(self.ctx, groups, done_cb)
        return self.add_pending(token)

    #
    #  General methods
    #
    def stream_sensor_data(self, interval: int, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.stream_sensor_data(self.ctx, interval, done_cb))

    def initialize_sensors(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.initialize_sensors(self.ctx, done_cb))

    def refresh_measurement_schedule(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.refresh_measurement_schedule(self.ctx, done_cb))

    def get_measurements(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.get_measurements(self.ctx, done_cb))

    def get_all_sensors(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.get_all_sensors(self.ctx, done_cb))

    def configure_measurement_schedule(
        self, measurements: Dict[str, str], done: DoneHWCommand = None
    ):
        service, done_cb = self.make_done(done)
        return self.add_pending(
            service.configure_measurement_schedule(self.ctx, measurements, done_cb)
        )

    def configure_temp_and_vccint(self, done: DoneHWCommand = None):  # pragma: no cover
        service, done_cb = self.make_done(done)
        return self.add_pending(service.configure_temp_and_vccint(self.ctx, done_cb))



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List

from chipscopy import dm
from chipscopy.dm import chipscope
from chipscopy.client import connect_xicom
from chipscopy.client.core_property_client import CorePropertyClient
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.proxies.AxisDDRProxy import SERVICE_NAME as AXIS_DDR_NAME


class AxisDDRClient(CorePropertyClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.type == "ddrmc"

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(AXIS_DDR_NAME)

    #
    #  General methods
    #
    def initialize(self, done: DoneHWCommand):
        service, done_cb = self.make_done(done)
        token = service.initialize(self.ctx, done_cb)
        return self.add_pending(token)

    def terminate(self, done: DoneHWCommand):
        service, done_cb = self.make_done(done)
        token = service.terminate(self.ctx, done_cb)
        return self.add_pending(token)

    def refresh_property(self, names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_property(self.ctx, names, done_cb)
        return self.add_pending(token)

    def refresh_cal_status(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_cal_status(self.ctx, done_cb)
        return self.add_pending(token)

    def refresh_cal_margin(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_cal_margin(self.ctx, done_cb)
        return self.add_pending(token)

    def get_cal_margin_mode(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.get_cal_margin_mode(self.ctx, done_cb)
        return self.add_pending(token)


def setup_debug_hub(cs_url: str, hw_url: str, hub_addrs: List[int]) -> Dict:
    server = connect_xicom(cs_url)
    server.connect_remote(hw_url)
    print(f"Connected to cs_server {cs_url} and hw_server {hw_url}")
    cs_service = server.get_sync_service("ChipScope")
    cs_view = server.get_view(chipscope)

    # Setup Debug Hub with addresses
    for node in cs_view.get_children():
        if "DPC" in node.props.get("Name"):
            print("Running Debug Hub setup")
            cs_service.setup_debug_cores(node.ctx, debug_hub_addrs=hub_addrs).get()
            print()

    print("ChipScope Nodes:")
    cs_view.print_tree(True)
    return cs_view


def get_sddr(view: Dict, mc_index=0) -> AxisDDRClient:
    ddrs = []
    # Find debug hub first and then initialize DDRs
    for node in list(view.get_all()):
        if node.type and node.type == "debug_hub":
            print("Debug Hub found.")
            ddrs.extend(view.find_nodes(node, AxisDDRClient))

    if len(ddrs):
        print("Soft DDR Core found and initialized")
        # Point to the target DDR by index
        ddr = ddrs[mc_index]
        return ddr
    else:
        return None



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Dict, List, Optional, Union

from chipscopy.api.containers import QueryList
from chipscopy.api.ibert.yk_scan import YKScan
from chipscopy.utils import printer

if TYPE_CHECKING:
    from chipscopy.api.ibert.rx import RX


UnionYKScanListYKScan = Union[YKScan, List[YKScan], QueryList[YKScan]]


class YKScanManager:
    """
    Factory class to manage YK scan creation and destruction
    """

    scans: ClassVar[Dict[str, YKScan]] = dict()
    last_scan_number: ClassVar[int] = 0
    scan_name_prefix: ClassVar[str] = "YKScan_"

    @staticmethod
    def all_yk_scans() -> QueryList[YKScan]:
        """
        Get all :py:class:`YKScan` object(s)

        Returns:
            List of YK scan object(s)

        """
        return QueryList(YKScanManager.scans.values())

    @staticmethod
    def create_yk_scans(*, target_objs: Union[RX, List[RX]]) -> QueryList[YKScan]:
        """
        Create an instance of :py:class:`YKScan` and attach it to the ``yk_scan`` attribute
        of the ``target_obj``(s)

        Args:
            target_objs: The object to use for attaching the YK scan instance.
                The object **must** be an instance of ``RX`` class.

        Returns:
            List of YK scan object(s) created

        """
        rxs = target_objs
        if not isinstance(target_objs, list):
            rxs = [target_objs]

        new_yk_scans = QueryList()

        rx_without_support = list()

        for rx in rxs:
            if not hasattr(rx, "yk_scan"):
                rx_without_support.append(rx.name)
                continue

            YKScanManager.last_scan_number += 1
            scan_name = f"{YKScanManager.scan_name_prefix}{YKScanManager.last_scan_number}"

            try:
                new_scan = YKScan(rx=rx, name=scan_name)
            except Exception as e:
                YKScanManager.last_scan_number -= 1
                raise e

            YKScanManager.scans[new_scan.name] = new_scan
            new_yk_scans.append(new_scan)

        if len(rx_without_support) > 0:
            rxs = "\n".join(rx_without_support)
            printer(f"YK scan creation isn't supported for following RX(s)\n{rxs}", level="warning")

        return new_yk_scans

    @staticmethod
    def delete_yk_scans(scans_to_delete: Optional[UnionYKScanListYKScan] = None):
        """
        Delete :py:class:`YKScan` object(s) that were created previously

        Args:
            scans_to_delete: **(Optional)** YK scan object(s) to delete.

        .. warning::
            If no YK scan objects are provided, all available YK scans are deleted.

        """
        if scans_to_delete is None:
            scans_to_delete = [eye_scan for eye_scan in YKScanManager.scans.values()]

        for eye_scan in scans_to_delete:
            name = eye_scan.name
            YKScanManager.scans[name].invalidate()
            del YKScanManager.scans[name]

        # If there are no scans left, reset the scan numbering
        if len(YKScanManager.scans) == 0:
            YKScanManager.last_scan_number = -1



# *****************************************************************************
# * Copyright (c) 2020 Xilinx, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************

from typing import Any, Dict, ByteString
from chipscopy.tcf.services import Service, DoneHWCommand
from chipscopy.tcf import services

XICOM_SERVICE = "Xicom"
XICOM_EVEREST_SERVICE = "XicomEverest"
"""Xicom services names."""


class DoneXicom(services.DoneHWCommand):
    """Client call back interface for generic commands."""
    def doneHW(self, token, error, results):
        """Called when memory operation command command is done.

        :param token: Command handle.
        :param error: Error object or **None**.
        :param results: List of results from command
        """
        pass


class XicomService(Service):
    """TCF Xicom service interface."""

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return XICOM_SERVICE

    def get_hw_server_version_info(self, done, **props):
        """ """
        raise NotImplementedError("Abstract method")

    def config_begin(self, ctx: str, props: Dict[str, Any] = {}, done: DoneHWCommand = None) -> services.TokenType:
        """
        Initiates the configuration process.

        :param ctx: Context ID of the device to configure
        :param props: Configuration properties
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        raise NotImplementedError("Abstract method")

    def config_data(self, ctx: str, data: ByteString, done: DoneHWCommand = None) -> services.TokenType:
        """
        Sends chunk of configuration data

        :param ctx: Context ID of the device to configure
        :param data: Data chunk to configure
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        raise NotImplementedError("Abstract Method")

    def config_end(self, ctx: str, props: Dict[str, Any] = {}, done: DoneHWCommand = None) -> services.TokenType:
        """
        Ends the configuration process.

        :param ctx: Context ID of the device to configure
        :param props: Configuration properties
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        raise NotImplementedError("Abstract method")

    def config_reset(self, ctx: str, props: Dict[str, Any] = {}, done: DoneHWCommand = None) -> services.TokenType:
        """
        Resets the device configuration.

        :param ctx: Context ID of the device to configure
        :param props: Configuration reset properties
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        raise NotImplementedError("Abstract method")

    def config_secure_debug(self, ctx: str, data: ByteString, done: DoneHWCommand = None) -> services.TokenType:
        """
        Enables secure debug using authentication data.

        :param ctx: Context ID of the device
        :param data: Authentication data
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        raise NotImplementedError("Abstract Method")

    def config_in(self, ctx, start_pos, total_bytes, data, done):
        raise NotImplementedError("Abstract method")

    def config_start(self, ctx, done):
        raise NotImplementedError("Abstract method")

    def jtag_reg_list(self, ctx: str or int, done: DoneHWCommand = None) -> services.TokenType:
        """
        Retrieves list of available JTAG register context IDs

        :param ctx: Context ID of device or idcode of a jtag device
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        raise NotImplementedError("Abstract method")

    def jtag_reg_def(self, reg_ctx: str, done: DoneHWCommand = None) -> services.TokenType:
        """
        Retrieves the definition of a given JTAG register context ID

        :param reg_ctx: JTAG register context ID
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        raise NotImplementedError("Abstract method")

    def jtag_reg_get(self, ctx: str, reg_ctx: str, done: DoneHWCommand = None) -> services.TokenType:
        """
        Retrieves the current value for a given register

        :param ctx: Context ID of device
        :param reg_ctx: JTAG register context ID
        :param done: Callback with the result and any error
        :return: Token of command sent
        """
        raise NotImplementedError("Abstract method")

class XicomEverestService(Service):
    """TCF JtagCable service interface."""

    def getName(self):
        """Get this service name.

        :returns: The value of string :const:`NAME`
        """
        return XICOM_EVEREST_SERVICE

    def config(self, ctx: str, data: bytes, options: Dict[str, Any] = {}, done: DoneHWCommand = None) -> Dict[str, Any]:
        """
        Configures the device with given PDI data

        :param ctx: Context ID of the device to update
        :param data: Binary data to configure device
        :param options: dict of optional parameters for configuration
        :param done: Callback with the result and any error
        :return: Dictionary of success
        """
        raise NotImplementedError("Abstract method")

    def get_status(self, ctx: str, need_definitions: bool, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Grabs the status registers from the pmc_tap
        :param ctx: Context ID of the device to update
        :param need_definitions: Flag to grab the bit definitions of the status registers
        :param done: Callback with the result and any error
        :return: Dictionary of status registers
        """
        raise NotImplementedError("Abstract method")

    def get_static(self, ctx: str, need_definitions: bool, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Grabs the static status registers from the pmc_tap
        :param ctx: Context ID of the device to update
        :param need_definitions: Flag to grab the bit definitions of the status registers
        :param done: Callback with the result and any error
        :return: Dictionary of status registers
        """
        raise NotImplementedError("Abstract method")

    def reset(self, ctx: str, done: DoneHWCommand) -> Dict[str, Any]:
        """
        Runs a system reset through the pmc_tap
        :param ctx: Context ID of the device to reset
        :param done: Callback with the result and any error
        :return: Dictionary of status registers
        """
        raise NotImplementedError("Abstract method")




# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ClassVar, Dict, List, Optional, Union

from chipscopy.api.containers import QueryList
from chipscopy.api.ibert.link import RX, TX, Link, LinkGroup

UnionLinkListLink = Union[Link, List[Link], QueryList[Link]]
UnionLinkGroupListLinkGroup = Union[LinkGroup, List[LinkGroup], QueryList[LinkGroup]]


class LinkManager:
    links: ClassVar[Dict[str, Link]] = dict()
    last_link_number: ClassVar[int] = -1
    link_name_prefix: ClassVar[str] = "Link_"

    @staticmethod
    def all_links() -> QueryList[Link]:
        """
        Get all :py:class:`~Link` object(s)

        Returns:
            List of link(s)
        """
        return QueryList(LinkManager.links.values())

    @staticmethod
    def create_links(*, rxs: Union[RX, List[RX]], txs: Union[TX, List[TX]]) -> QueryList[Link]:
        """
        Create link(s) for given :py:class:`~RX` object(s) and :py:class:`~TX` object(s).
        None of the RX or TX object(s) must be part of existing links.

        Args:
            rxs: RX object(s) to use for link creation

            txs: TX object(s) to use for link creation

            .. NOTE:: If passing a list of RX objects and TX objects, the length of both the lists
                must be same for link creation to succeed. For link creation, RX and TX objects
                are used in the order they appear in the lists.

        Returns:
            Newly created :py:class:`Link` object(s)

        """
        # Sanitize the input
        if isinstance(rxs, list) and isinstance(txs, list):
            if len(rxs) != len(txs):
                raise ValueError("Number of RXs and TXs provided in the lists must be the same!")

            for index, pair in enumerate(zip(rxs, txs)):
                rx, tx = pair
                if isinstance(rx, str):
                    rxs[index] = None
                if isinstance(tx, str):
                    txs[index] = None

        elif isinstance(rxs, RX) and isinstance(txs, TX):
            rxs = [rxs]
            txs = [txs]

        elif isinstance(rxs, str) and isinstance(txs, TX):
            rxs = [None]
            txs = [txs]

        elif isinstance(rxs, RX) and isinstance(txs, str):
            rxs = [rxs]
            txs = [None]

        else:
            raise ValueError(
                "Valid formats for link creation are\n"
                "- rxs = RX obj, txs = TX obj\n"
                "- rxs = list[RX objs], txs = list[TX objs]; Both the lists must equal in length\n"
                "- Either rxs/txs = 'Unknown', txs/rxs = TX/RX obj\n"
            )

        new_links = QueryList()

        for rx, tx in zip(rxs, txs):
            LinkManager.last_link_number += 1
            link_name = f"{LinkManager.link_name_prefix}{LinkManager.last_link_number}"

            try:
                new_links.append(Link(rx, tx, link_name))
            except Exception as e:
                # IF creation fails, decrement by 1 since we didn't create a link
                LinkManager.last_link_number -= 1
                raise e

        LinkManager.links.update({link.name: link for link in new_links})

        return new_links

    @staticmethod
    def delete_links(links_to_delete: Optional[UnionLinkListLink] = None):
        """
        Delete links that were created previously

        Args:
            links_to_delete: **(Optional)** Link objects to delete.
                .. NOTE::If no link objects are provided, all available links are deleted.
        """
        if links_to_delete is None:
            links_to_delete = [link for link in LinkManager.links.values()]

        if isinstance(links_to_delete, Link):
            links_to_delete = [links_to_delete]

        for link in links_to_delete:
            name = link.name
            LinkManager.links[name].invalidate()
            del LinkManager.links[name]

        # If there are no links left, reset the link numbering
        if len(LinkManager.links) == 0:
            LinkManager.last_link_number = -1


class LinkGroupManager:
    link_groups: ClassVar[Dict[str, LinkGroup]] = dict()
    last_link_group_number: ClassVar[int] = -1
    link_group_name_prefix: ClassVar[str] = "LinkGroup_"

    @staticmethod
    def all_link_groups() -> QueryList[LinkGroup]:
        """
        Get all :py:class:`~LinkGroup` object(s)

        Returns:
            List of link groups(s)
        """
        return QueryList(LinkGroupManager.link_groups.values())

    @staticmethod
    def create_link_groups(descriptions: Union[str, List[str]]) -> QueryList[LinkGroup]:
        """
        Create a link group for with description

        Args:
            descriptions: Description(s) for the link group(s) to create.

        Returns:
            Newly created link groups.

        """
        if isinstance(descriptions, str):
            descriptions = [descriptions]

        new_link_groups = QueryList()

        for description in descriptions:
            LinkGroupManager.last_link_group_number += 1
            group_name = f"{LinkGroupManager.link_group_name_prefix}{LinkGroupManager.last_link_group_number}"

            try:
                new_group = LinkGroup(group_name, description, dict())
            except Exception as e:
                LinkGroupManager.last_link_group_number -= 1
                raise e

            new_link_groups.append(new_group)

        LinkGroupManager.link_groups.update({group.name: group for group in new_link_groups})

        return new_link_groups

    @staticmethod
    def delete_link_groups(
        link_groups: UnionLinkGroupListLinkGroup, *, delete_links_in_group: bool = False
    ):
        """
        Delete link group(s)

        Args:
            link_groups: The link groups to delete

            delete_links_in_group: If set to ``True`` the links within each link group
                are deleted as well. Default behavior is to disassociate the link from the
                link group and not delete the link itself.

        Returns:

        """
        if isinstance(link_groups, LinkGroup):
            link_groups = [link_groups]

        links_to_delete = list()

        for group in link_groups:
            if delete_links_in_group:
                links_to_delete.extend(group.links)

            group_name = group.name
            # This will set the link_group field in the links to None
            group.invalidate()
            del LinkGroupManager.link_groups[group_name]

        # If there are no link groups left, reset the group numbering
        if len(LinkGroupManager.link_groups) == 0:
            LinkGroupManager.last_link_group_number = -1

        LinkManager.delete_links(links_to_delete)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from collections import defaultdict
from queue import Queue
from typing import (
    Generic,
    Iterable,
    Optional,
    Callable,
    TYPE_CHECKING,
    Set,
    List,
    Dict,
    Any,
    Union,
    Tuple,
    TypeVar,
)
from threading import Lock
from datetime import datetime
from dataclasses import dataclass

from chipscopy.utils.printer import printer

if TYPE_CHECKING:
    from chipscopy.dm import Node


T = TypeVar("T")


@dataclass
class PropertyUpdateEvent:
    names: List[str]
    """Names of properties in this update"""
    values: List[Any]
    """New value of properties"""
    timestamp: datetime
    """Timestamp the update event was received"""

    def __post_init__(self):
        self._max_length = 0
        for name in self.names:
            if len(name) > self._max_length:
                self._max_length = len(name)

    def __repr__(self) -> str:
        retval = f"PropertyUpdateEvent\n\tTime stamp - {str(self.timestamp)}\n"
        for name, value in zip(self.names, self.values):
            retval += f"\t{name}{' ' * (self._max_length - len(name))} = {value}\n"

        return retval


WatchlistEventHandler = Callable[["Queue[PropertyUpdateEvent]"], None]


class Watchlist(Generic[T]):
    def __init__(self, parent: T):
        self.parent = parent
        self.core_tcf_node = self.parent.core_tcf_node

        self._listeners: Dict[str, List[WatchlistEventHandler]] = defaultdict(list)
        self._update_buckets: Dict[str, "Queue[PropertyUpdateEvent]"] = defaultdict(Queue)
        self._watch_ids_for_property: Dict[str, Set[str]] = defaultdict(set)
        self._tcf_event_listener_registered: bool = False

        self._lock = Lock()

    @property
    def active_properties(self):
        return set(self._watch_ids_for_property)

    def _get_uuid(self):
        watch_id = str(uuid.uuid4())
        while watch_id in self._listeners:
            watch_id = str(uuid.uuid4())

        return watch_id

    def add(
        self,
        property_names: Union[str, List[str]],
        *,
        listeners: Optional[Union[WatchlistEventHandler, List[WatchlistEventHandler]]] = None,
    ):
        if listeners is not None and not isinstance(listeners, list):
            listeners = [listeners]

        sanitized_data = PropertyCommands.sanitize_input(property_names, list)

        self.core_tcf_node.add_to_property_watchlist(sanitized_data)

        with self._lock:
            if not self._tcf_event_listener_registered:
                self.parent.core_tcf_node.add_listener(self._mandatory_event_listener)
                self._tcf_event_listener_registered = True

            watch_id = self._get_uuid()
            self._listeners[watch_id].extend(listeners)

            for property_name in sanitized_data:
                self._watch_ids_for_property[property_name].add(watch_id)

    def remove(self, property_names: Union[str, List[str]]):
        sanitized_data = {
            prop
            for prop in PropertyCommands.sanitize_input(property_names, list)
            if prop in self.active_properties
        }

        self.core_tcf_node.remove_from_property_watchlist(sanitized_data)

        with self._lock:
            for prop in sanitized_data:
                del self._watch_ids_for_property[prop]

    def _mandatory_event_listener(self, node: "Node", updated_properties: Set[str]):
        # NOTE - This is called on the TCF event dispatcher thread

        updated_properties = list(updated_properties)
        updated_properties.sort()

        updates_by_id = dict()

        for property_name in updated_properties:
            with self._lock:
                if property_name not in self._watch_ids_for_property:
                    continue

                for watch_id in self._watch_ids_for_property[property_name]:
                    if watch_id not in updates_by_id:
                        # list at index 0 will hold the property names
                        # list at index 1 will hold property names
                        updates_by_id[watch_id]: Tuple[List[str], List[Any]] = (list(), list())

                    updates_by_id[watch_id][0].append(property_name)
                    updates_by_id[watch_id][1].append(node.props[property_name])

        if len(updates_by_id) == 0:
            return

        time_stamp = datetime.now()
        for watch_id, updates in updates_by_id.items():
            names, values = updates
            self._update_buckets[watch_id].put(
                PropertyUpdateEvent(names=names, values=values, timestamp=time_stamp)
            )

        # Call the event_listener registered by the user
        for watch_id in updates_by_id:
            with self._lock:
                if watch_id not in self._listeners or self._listeners[watch_id] is None:
                    continue

                for listener in self._listeners[watch_id]:
                    try:
                        listener(self._update_buckets[watch_id])
                    except Exception as e:
                        printer(
                            f"Unhandled exception when calling listener for watch ID {watch_id}"
                            f"\nException - {str(e)}",
                            level="warning",
                        )


class PropertyCommands(Generic[T]):
    def __init__(self, parent: T):
        self.parent = parent
        self.core_tcf_node = self.parent.core_tcf_node

        self.watchlist: Watchlist["PropertyCommands"] = Watchlist(self)

    @staticmethod
    def sanitize_input(user_input, desired_format):
        if desired_format == list:
            if isinstance(user_input, str):
                return [user_input]
            elif isinstance(user_input, Iterable):
                return list(user_input)

        elif desired_format == dict:
            if isinstance(user_input, dict):
                return user_input

        raise TypeError(f"Unsupported type {type(user_input)}!")

    def get(self, property_names: Union[str, List[str]]):
        return self.core_tcf_node.get_property(
            PropertyCommands.sanitize_input(property_names, list)
        )

    def set(self, **property_dict):
        return self.core_tcf_node.set_property(PropertyCommands.sanitize_input(property_dict, dict))

    def refresh(self, property_names: Union[str, List[str]]):
        return self.core_tcf_node.refresh_property(
            PropertyCommands.sanitize_input(property_names, list)
        )

    def commit(self, property_names: Union[str, List[str]]):
        return self.core_tcf_node.commit_property(
            PropertyCommands.sanitize_input(property_names, list)
        )

    def report(self, property_names: Union[str, List[str]]):
        return self.core_tcf_node.report_property(
            PropertyCommands.sanitize_input(property_names, list)
        )



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
from chipscopy.tcf.services import Service, DoneHWCommand
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy

SYSMON_SERVICE_NAME = "SysMon"


class SysMonService(Service):
    def getName(self) -> str:
        return SYSMON_SERVICE_NAME


class SysMonProxy(CorePropertyProxy, SysMonService):
    def __init__(self, channel):
        super(SysMonProxy, self).__init__(channel)
        self.listeners = {}

    def stream_sensor_data(self, node_id: str, interval: int, done: DoneHWCommand) -> None:
        return self.send_xicom_command("streamSensorData", (node_id, interval), done)

    def initialize_sensors(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("initializeSensors", (node_id,), done)

    def refresh_measurement_schedule(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("refreshMeasurementSchedule", (node_id,), done)

    def get_measurements(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("getMeasurements", (node_id,), done)

    def get_all_sensors(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_xicom_command("getAllSensors", (node_id,), done)

    def configure_measurement_schedule(
        self, node_id: str, measurements: Dict[str, str], done: DoneHWCommand
    ):
        return self.send_xicom_command(
            "configureMeasurementSchedule", (node_id, measurements), done
        )

    def configure_temp_and_vccint(self, node_id: str, done: DoneHWCommand):  # pragma: no cover
        return self.send_xicom_command("configureTempAndVCCINT", (node_id,), done)



# %% [markdown]
# <link rel="preconnect" href="https://fonts.gstatic.com">
# <link href="https://fonts.googleapis.com/css2?family=Fira+Code&display=swap" rel="stylesheet">
#
# ### License
#
# <p style="font-family: 'Fira Code', monospace; font-size: 1.2rem">
# Copyright 2021 Xilinx, Inc.<br><br>
# Licensed under the Apache License, Version 2.0 (the "License");<br>
# you may not use this file except in compliance with the License.<br><br>
# You may obtain a copy of the License at <a href="http://www.apache.org/licenses/LICENSE-2.0"?>http://www.apache.org/licenses/LICENSE-2.0</a><br><br>
# Unless required by applicable law or agreed to in writing, software<br>
# distributed under the License is distributed on an "AS IS" BASIS,<br>
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.<br>
# See the License for the specific language governing permissions and<br>
# limitations under the License.<br>
# </p>
#

# %% [markdown]
# # ChipScoPy Device Programming Example
#
#
# <img src="../img/api_overview.png" width="500" align="left">

# %% [markdown]
# ## Description
# This example demonstrates how to program a device using the ChipScoPy Python API.
#
#
# ## Requirements
# - Local or remote Xilinx Versal board, such as a VCK190
# - Xilinx hw_server 2021.2 installed and running
# - Python 3.8 or greater installed
# - ChipScoPy 2021.2 installed
# - Jupyter notebook support installed - Please do so, using the command `pip install chipscopy[jupyter]`

# %% [markdown]
# ## 1 - Initialization: Imports and File Paths
#
# After this step,
# - Required functions and classes are imported
# - Paths to server(s) and files are set correctly

# %%
import os
from chipscopy import get_design_files
from chipscopy import create_session, report_versions

# %%
# Make sure to start the hw_server prior to running.
# Specify location of the running hw_server below.
# The default is localhost - but can be other locations on the network.
HW_URL = os.getenv("HW_SERVER_URL", "TCP:localhost:3121")

# The get_design_files() function tries to find the programming and probes
# files for an included example design.
PROGRAMMING_FILE = get_design_files("vck190/production/chipscopy_ced").programming_file

print(f"HW_URL: {HW_URL}")
print(f"PROGRAMMING_FILE: {PROGRAMMING_FILE}")


# %% [markdown]
# ## 2 - Create a session and connect to the hw_server
#
# The session is a container that keeps track of devices and debug cores.
# After this step,
# - Session is initialized and connected to server(s)
# - Versions are detected and reported to stdout
#
# *NOTE*: No cs_server is required for this example.

# %%
session = create_session(hw_server_url=HW_URL)
report_versions(session)

# %% [markdown]
# ## 3 - Program the device with the example design
#
# After this step,
# - Device is programmed with the example programming file

# %%
# Typical case - one device on the board - get it.
device = session.devices.filter_by(family="versal").get()
device.program(PROGRAMMING_FILE)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing_extensions import Final

TYPE: Final[str] = "Type"
CHILDREN: Final[str] = "Children"
ALIAS_DICT: Final[str] = "Alias dict"
HANDLE_NAME: Final[str] = "Handle name"
PROPERTY_ENDPOINT: Final[str] = "Property endpoint"
MODIFIABLE_ALIASES: Final[str] = "Modifiable aliases"

COUNT_KEY: Final[str] = "Count"
PROPERTIES_KEY: Final[str] = "Properties"
NAME_PREFIX_KEY: Final[str] = "Name prefix"

# -------------------------------------------------------------------
# Serial object types
# -------------------------------------------------------------------
GT_KEY: Final[str] = "GT"
"""GT"""
TX_KEY: Final[str] = "TX"
"""TX"""
RX_KEY: Final[str] = "RX"
"""RX"""
PLL_KEY: Final[str] = "PLL"
"""PLL"""
IBERT_KEY: Final[str] = "IBERT"
"""IBERT"""
GT_GROUP_KEY: Final[str] = "GT_Group"
"""GT Group"""
MICROBLAZE_KEY: Final[str] = "MicroBlaze"
"""MicroBlaze"""
PLL_TYPE_KEY: Final[str] = "PLL_Type"
"""PLL"""

# ----------------------------------------------------------------
# Applicable to both RX and TX
# ----------------------------------------------------------------
PATTERN: Final[str] = "Pattern"
PLL_SOURCE: Final[str] = "PLL Source"
PLL_LOCK_STATUS: Final[str] = "PLL Lock Status"

# ----------------------------------------------------------------
# Applicable to TX only
# ----------------------------------------------------------------
TX_RESET: Final[str] = "Reset"

# ----------------------------------------------------------------
# Applicable to RX only
# ----------------------------------------------------------------
RX_BER: Final[str] = "BER"
RX_RESET: Final[str] = "Reset"
RX_STATUS: Final[str] = "Status"
RX_YK_SCAN: Final[str] = "YK scan"
RX_EYE_SCAN: Final[str] = "Eye scan"
RX_LOOPBACK: Final[str] = "Loopback"
RX_LINE_RATE: Final[str] = "Line Rate"
RX_SUPPORTED_SCANS: Final[str] = "Supported scans"
RX_RECEIVED_BIT_COUNT: Final[str] = "Received Bit Count"
RX_PATTERN_CHECKER_ERROR_COUNT: Final[str] = "Pattern Checker Error Count"

# ----------------------
# Microblaze
# ----------------------
MB_ELF_VERSION: Final[str] = "ELF Version"

# -----------------------------------
# Eye scan related
# -----------------------------------
EYE_SCAN_DONE: Final[str] = "Done"
"""Eye scan completed successfully"""
EYE_SCAN_ABORTED: Final[str] = "Aborted"
"""Eye scan was aborted by user or cs_server"""
EYE_SCAN_NOT_STARTED: Final[str] = "Not Started"
"""Eye scan object created, but scan not started in MB"""
EYE_SCAN_IN_PROGRESS: Final[str] = "In Progress"
"""Eye scan is in progress"""

EYE_SCAN_HORZ_STEP: Final[str] = "Horizontal Step"
"""Alias for horizontal step eye scan parameter"""
EYE_SCAN_VERT_STEP: Final[str] = "Vertical Step"
"""Alias for vertical step eye scan parameter"""
EYE_SCAN_HORZ_RANGE: Final[str] = "Horizontal Range"
"""Alias for horizontal range eye scan parameter"""
EYE_SCAN_VERT_RANGE: Final[str] = "Vertical Range"
"""Alias for horizontal range eye scan parameter"""
EYE_SCAN_TARGET_BER: Final[str] = "Target BER"
"""Alias for target BER eye scan parameter"""
EYE_SCAN_DWELL_TIME: Final[str] = "Dwell Time"
"""Alias for dwell time eye scan parameter"""

YK_SCAN_SNR_VALUE: Final[str] = "YK Scan SNR Value"
YK_SCAN_STOP_TIME: Final[str] = "YK Scan Stop Time"
YK_SCAN_START_TIME: Final[str] = "YK Scan Start Time"
YK_SCAN_SLICER_DATA: Final[str] = "YK Scan Slicer Data"

EYE_SCAN_UT: Final[str] = "UT"
EYE_SCAN_STATUS: Final[str] = "Status"
EYE_SCAN_2D_PLOT: Final[str] = "2D Plot"
EYE_SCAN_RAW_DATA: Final[str] = "Raw Scan Data"
EYE_SCAN_PROGRESS: Final[str] = "Scan Progress"
EYE_SCAN_PRESCALE: Final[str] = "Prescale"
EYE_SCAN_STOP_TIME: Final[str] = "Eye Scan Stop Time"
EYE_SCAN_START_TIME: Final[str] = "Eye Scan Start Time"
EYE_SCAN_ERROR_COUNT: Final[str] = "Error Count"
EYE_SCAN_SAMPLE_COUNT: Final[str] = "Sample Count"
EYE_SCAN_2D_PLOT_DATA: Final[str] = "Data"
EYE_SCAN_SCAN_PARAMETERS: Final[str] = "Scan Parameters"
EYE_SCAN_2D_PLOT_BER_FLOOR_VALUE: Final[str] = "BER Floor Value"
EYE_SCAN_TOTAL_NO_OF_DATA_POINTS_READ: Final[str] = "Total #Data Points Read"
EYE_SCAN_TOTAL_NO_OF_DATA_POINTS_EXPECTED: Final[str] = "Total #Data Points Expected"

# ----------------------------------------------------------------
# PLL related
# ----------------------------------------------------------------
PLL_RESET: Final[str] = "Reset"



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

"""
Simple interactive shell for TCF.  This is basically a Python interpreter with
a few TCF extensions.

Usage:
    python -m tcf.shell

Commands:
    peers              - Print discovered peers
    connect(params)    - Connect to TCF peer, params =
                         "<protocol>:<host>:<port>"
    cmd.<service>.<command<(args)
                       - Send command to remote service and return result
    disconnect         - Disconnect from peer
    events.record(<service>)
                       - Start recording events for service
    events             - Print last recorded events
    events.stop([<service>])
                       - Stop recording for service or for all services
"""

import code
import os
import sys


try:
    import tcf
except ImportError:
    # add parent dir to path
    sys.path.insert(0, os.path.dirname(os.getcwd()))
    import tcf

from tcf.util import sync, event  # @UnresolvedImport
from tcf import protocol, channel  # @UnresolvedImport


class print_peers:
    "Print list of discovered peers"
    def __call__(self):
        return tcf.peers()

    def __repr__(self):
        peers = tcf.peers()
        return '\n'.join(["%s, %s" % (p.getID(), p.getName())
                          for p in list(peers.values())])


class Shell(code.InteractiveConsole, protocol.ChannelOpenListener,
            channel.ChannelListener):
    def __init__(self):
        locals = {  # @ReservedAssignment
            "connect": tcf.connect,
            "peers": print_peers()
        }
        protocol.startEventQueue()
        protocol.startDiscovery()
        protocol.invokeAndWait(protocol.addChannelOpenListener, self)
        code.InteractiveConsole.__init__(self, locals)

    def interact(self, banner=None):
        try:
            try:
                ps1 = sys.ps1  # @UndefinedVariable
            except AttributeError:
                ps1 = None
            sys.ps1 = "tcf> "
            super(Shell, self).interact(banner)
        finally:
            if ps1:
                sys.ps1 = ps1
            else:
                del sys.ps1
            protocol.invokeLater(protocol.removeChannelOpenListener, self)
            protocol.shutdownDiscovery()
            protocol.getEventQueue().shutdown()

    def onChannelOpen(self, channel):
        wrapper = sync.DispatchWrapper(channel)
        self.locals["channel"] = wrapper
        self.locals["disconnect"] = wrapper.close
        self.locals["cmd"] = sync.CommandControl(channel, interactive=True)
        self.locals["events"] = event.EventRecorder(channel)
        protocol.invokeAndWait(protocol.removeChannelOpenListener, self)
        wrapper.addChannelListener(self)

    def onChannelClosed(self, error):
        del self.locals["channel"]
        del self.locals["cmd"]
        del self.locals["disconnect"]
        del self.locals["events"]
        protocol.addChannelOpenListener(self)


def interact():
    shell = Shell()
    try:
        # enable commandline editing if available
        import readline
    except ImportError:
        pass
    else:
        import rlcompleter
        readline.set_completer(rlcompleter.Completer(shell.locals).complete)
        readline.parse_and_bind("tab: complete")
    shell.interact("TCF Shell")

if __name__ == "__main__":
    interact()



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from collections import namedtuple
from typing import Optional, List

from chipscopy.api.device.device_identification import (
    _DeviceIdentificationTuple,
    _DeviceNodeIdentificationTuple,
    DeviceIdentification,
)
from chipscopy.client import ServerInfo
from chipscopy.client.jtagdevice import JtagDevice, JtagCable, JtagRegister
from chipscopy.dm import Node
from chipscopy.client.view_info import ViewInfo


def get_device_dna(node: Node) -> Optional[int]:
    """Returns: 128-bit device dna value if available for the node, None otherwise."""
    dna_128 = None
    if node.props.get("node_cls") is JtagDevice:
        # jtag devices store dna as a bytestream
        if node.props.get("regs", None):
            if node.props["regs"].get("dna", None):
                # Below refreshes the dna data and makes it available
                # node.status('dna')
                node.update_regs(reg_names=("dna",), force=True, done=None)
                jtag_register = node.props["regs"]["dna"]
                bytearray_data = jtag_register.data
                dna_128 = int.from_bytes(bytearray_data, byteorder="little", signed=False)
    else:
        # Other contexts store dna as a 4-tuple of 32-bit ints
        dna_4_tuple = node.props.get("DeviceDNA", None)
        if dna_4_tuple and len(dna_4_tuple) == 4:
            dna_128 = (
                dna_4_tuple[3] << 96 | dna_4_tuple[2] << 64 | dna_4_tuple[1] << 32 | dna_4_tuple[0]
            )
    return dna_128


def get_jtag_device_for_dna(hw_server: ServerInfo, dna_128: int) -> Optional[namedtuple]:
    JtagDeviceInfo = namedtuple(
        "JtagDeviceInfo", "cable_ctx cable_name chain_index node_ctx family"
    )
    view = hw_server.get_view("jtag")
    for node in view.get_children():
        # Jtag cables live right under the jtag view.
        jtag_cable_node = view.get_node(node.ctx, JtagCable)
        if jtag_cable_node:
            jtag_index = 0
            for child in view.get_children(parent=jtag_cable_node.ctx):
                jtag_device_node = view.get_node(child.ctx, JtagDevice)
                # I ass-u-me the node order is the same order as the physical jtag chain.
                if jtag_device_node:
                    dna_128_ = get_device_dna(jtag_device_node)
                    if dna_128_ == dna_128:
                        family = jtag_device_node.props.get("arch_name", None)
                        jtag_cable_name = jtag_cable_node.ctx
                        jtag_cable_desc = jtag_cable_node.props.get("Description", None)
                        jtag_cable_serial = jtag_cable_node.props.get("Serial", None)
                        if jtag_cable_desc:
                            jtag_cable_name = jtag_cable_desc
                            if jtag_cable_serial:
                                jtag_cable_name += " " + jtag_cable_serial
                        return JtagDeviceInfo(
                            jtag_cable_node.ctx,
                            jtag_cable_name,
                            jtag_index,
                            jtag_device_node.ctx,
                            family,
                        )
                    jtag_index += 1
    return None


def upgrade_node(view: ViewInfo, node: Node) -> Node:
    cls = node.props.get("node_cls")
    if cls.is_compatible(node):
        # Upgrade if possible
        node = view.get_node(node.ctx, cls)
    return node


def _get_node_hier_name(view: ViewInfo, node: Node) -> str:
    name = node.props["Name"]
    while node.parent_ctx:
        prev_node = node
        node = view.get_node(node.parent_ctx)
        if node == prev_node:
            break
        parent_name = node.props["Name"]
        name = f"{parent_name}/{name}"
    return name


def _copy_node_props(node: Node, props: dict) -> dict:
    transformed_props = {}
    for k, v in props.items():
        if type(v) == int or type(v) == str or type(v) == bool:
            transformed_props[str(k)] = v
        elif type(v) == JtagRegister:
            node.update_regs(reg_names=(f"{k}",), force=True, done=None)
            int_val = int.from_bytes(v.data, byteorder="little", signed=False)
            transformed_props[str(k)] = int_val
        elif type(v) == dict:
            transformed_props[str(k)] = _copy_node_props(node, v)
        elif type(v) == bytearray:
            int_val = int.from_bytes(v, byteorder="little", signed=False)
            transformed_props[str(k)] = int_val
        else:
            transformed_props[str(k)] = str(v)
    return transformed_props


def _create_jtag_node_dict(
    server_name: str,
    view_name: str,
    jtag_cable_ctx: str,
    jtag_index: int,
    view: ViewInfo,
    node: Node,
) -> dict:
    children = list(view.get_children(node))
    jtag_node_dict = {
        "server": server_name,
        "view": view_name,
        "ctx": node.ctx,
        "dna": get_device_dna(node),
        "family": node.props.get("arch_name", None),
        "name": node.props["Name"],
        "hier_name": _get_node_hier_name(view, node),
        "jtag_cable_ctx": jtag_cable_ctx,
        "jtag_index": jtag_index,
        "props": _copy_node_props(node, node.props),
        "parent_ctx": node["ParentID"],
        "children": {
            child.ctx: _create_jtag_node_dict(
                server_name, view_name, jtag_cable_ctx, jtag_index, view, child
            )
            for child in children
        },
    }
    return jtag_node_dict


def get_jtag_view_dict(hw_server: ServerInfo) -> dict:
    jtag_dict = {}
    server_name = "hw_server"
    view_name = "jtag"
    view = hw_server.get_view(view_name)
    for jtag_cable in view.get_children():
        device_dict = {}
        jtag_cable_ctx = jtag_cable.ctx
        device_nodes_to_process = list(view.get_children(jtag_cable))
        jtag_index = 0
        for node in device_nodes_to_process:
            device_dict[node.ctx] = _create_jtag_node_dict(
                server_name, view_name, jtag_cable_ctx, jtag_index, view, node
            )
        jtag_dict[jtag_cable_ctx] = {
            "name": jtag_cable.props["Name"],
            "ctx": jtag_cable_ctx,
            "devices": device_dict,
            "props": _copy_node_props(jtag_cable, jtag_cable.props),
        }
    return jtag_dict


def _create_basic_node_dict(server_name: str, view_name: str, view: ViewInfo, node: Node) -> dict:
    children = list(view.get_children(node))
    debugcore_node_dict = {
        "server": server_name,
        "view": view_name,
        "ctx": node.ctx,
        "dna": get_device_dna(node),
        "name": node.props["Name"],
        "hier_name": _get_node_hier_name(view, node),
        "props": _copy_node_props(node, node.props),
        "children": {
            child.ctx: _create_basic_node_dict(server_name, view_name, view, child)
            for child in children
        },
    }
    return debugcore_node_dict


def _get_basic_view_dict(server: ServerInfo, server_name, view_name) -> dict:
    device_dict = {}
    view = server.get_view(view_name)
    for node in view.get_children():
        device_dict[node.ctx] = _create_basic_node_dict(server_name, view_name, view, node)
    return device_dict


def get_memory_view_dict(hw_server: ServerInfo) -> dict:
    return _get_basic_view_dict(hw_server, "hw_server", "memory")


def get_debugcore_view_dict(hw_server: ServerInfo) -> dict:
    return _get_basic_view_dict(hw_server, "hw_server", "debugcore")


def get_chipscope_view_dict(cs_server: ServerInfo) -> dict:
    return _get_basic_view_dict(cs_server, "cs_server", "chipscope")


class DeviceScanner:
    """This is a utility to scan for live devices in hardware. It is used by the Session when creating the
    initial list of devices or rescanning hardware.

    scan_devices() generates a convenient list of DeviceIdentification named tuples based on the hw_server and
        cs_server connections.
    """

    def __init__(self, hw_server, cs_server):
        """
        Args:
            hw_server: hardware server
            cs_server: chipscope server(optional)
        """
        self.hw_server = hw_server
        self.cs_server = cs_server
        self.device_scan_results: Optional[List[DeviceIdentification]] = None

    def __repr__(self):
        device_identification_list = self.get_scan_results()
        return device_identification_list.__repr__()

    def __str__(self):
        raw_json = self.__repr__()
        parsed = json.loads(raw_json)
        retval = json.dumps(parsed, indent=4, sort_keys=True)
        return retval

    def __iter__(self):
        return self.device_scan_results.__iter__()

    def get_scan_results(self) -> List[DeviceIdentification]:
        if not self.device_scan_results:
            self.scan_devices()
        return self.device_scan_results

    @staticmethod
    def _get_children(server_type, view_name, view, node_hier_name, node) -> list:
        retval = list()
        for child in view.get_children(node):
            name_with_hierarchy = node_hier_name + "/" + child.props.get("Name")
            child_node_id = _DeviceNodeIdentificationTuple(
                server_type, view_name, child.props.get("Name"), child.ctx, name_with_hierarchy
            )
            retval.append(child_node_id)
            children = DeviceScanner._get_children(
                server_type, view_name, view, name_with_hierarchy, child
            )
            if children:
                retval.extend(children)
        return retval

    def scan_devices(self):
        """Query the hw_server for connected devices. Creates a list of devices found and connects together
        the associated contexts across jtag, memory, debugcore, and chipscope views.
        This is a convenience for finding and organizing nodes that may be living in unexpected places
        like DPC and Versal both living at the top level of a device. We figure it out here so the user
        doesn't have to.

        Upon successful scan completion, self.device_scan_results is populated with the scan results.
        """
        dna_values_already_found = {}
        # Use the debug core view to figure out how many devices we have
        devices_found = []
        views_to_add = list()
        views_to_add.append(("hw_server", "jtag", self.hw_server.get_view("jtag")))
        views_to_add.append(("hw_server", "memory", self.hw_server.get_view("memory")))
        views_to_add.append(("hw_server", "debugcore", self.hw_server.get_view("debugcore")))
        if self.cs_server:
            views_to_add.append(("cs_server", "chipscope", self.cs_server.get_view("chipscope")))
        for (server_type, view_name, view) in views_to_add:
            # Versal debugcore will currently have 2 top level targets for a single device -
            #    "Versal" and "DPC". This is bad - but both have a common DNA value so we can connect things together
            #    based on dna
            for node in view.get_all():
                if not node.props.get("Name"):
                    continue
                dna_128 = get_device_dna(node)
                node_hier_name = f"{view_name}/{node.props.get('Name')}"
                node_id = _DeviceNodeIdentificationTuple(
                    server_type=server_type,
                    view_name=view_name,
                    name=node.props.get("Name"),
                    context=node.ctx,
                    hier_name=node_hier_name,
                )
                node_id_list = [node_id]
                children = DeviceScanner._get_children(
                    server_type, view_name, view, node_hier_name, node
                )
                node_id_list.extend(children)
                if dna_128:
                    # Found device dna - If this is a versal device, make sure
                    # we associate the DPC and Versal top targets together
                    if dna_values_already_found.get(dna_128) is None:
                        jtag_device_info = get_jtag_device_for_dna(self.hw_server, dna_128)
                        if jtag_device_info:
                            # NOTE: it is possible that a single cs_server has multiple hw_server connections.
                            # In that case, jtag_device_info is None for the other hw_server nodes.
                            # We only add cs_server nodes for the correct hw_server.
                            jtag_cable_context = jtag_device_info.cable_ctx
                            jtag_cable_name = jtag_device_info.cable_name
                            jtag_index = jtag_device_info.chain_index
                            family = jtag_device_info.family
                            device_id = _DeviceIdentificationTuple(
                                family=family,
                                node_identification=node_id_list,
                                dna=dna_128,
                                cable_context=jtag_cable_context,
                                cable_name=jtag_cable_name,
                                jtag_index=jtag_index,
                            )
                            dna_values_already_found[dna_128] = device_id
                            devices_found.append(device_id)
                    else:
                        # Duplicate entry - just add the name we found to the list. Happens for Versal and DPC
                        dna_values_already_found[dna_128].node_identification.extend(node_id_list)
                else:
                    # No device dna on a node. Lots of reasons but this is not a node to worry about now.
                    # In the future we probably need to consider how to associate nodes with no dna.
                    #
                    name = node.props.get("Name")
                    if name.startswith("XVC:"):
                        # This is an XVC cable. Node name is in the format: "XVC:<host>:port"
                        # Use node name as a key instead of DNA for XVC cables to group things together
                        if dna_values_already_found.get(name) is None:
                            device_id = _DeviceIdentificationTuple(
                                family="xvc",
                                node_identification=node_id_list,
                                dna=None,
                                cable_context=None,
                                cable_name=name,
                                jtag_index=None,
                            )
                            dna_values_already_found[name] = device_id
                            devices_found.append(device_id)
                        else:
                            # Duplicate entry - just add the name we found to the list. Happens for Versal and DPC
                            dna_values_already_found[name].node_identification.extend(node_id_list)
                    else:
                        # print(f"*** IGNORING NODE: {node.props.get('Name')} - {node}")
                        pass

        # Keep in a somewhat consistent sorted order by jtag cable then jtag index if possible
        devices_found.sort(key=lambda x: [x.cable_name, x.jtag_index])
        self.device_scan_results = list()
        for device_tracker_tuple in devices_found:
            device_tracker = DeviceIdentification(device_tracker_tuple)
            self.device_scan_results.append(device_tracker)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from chipscopy.utils.logger import log
from chipscopy import dm
from chipscopy.client import core_property_client
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.proxies.AxisVIOProxy import AXIS_VIO_SERVICE_NAME, AXIS_VIO_NODE_NAME

# NOTE: PLEASE USE THIS DOMAIN NAME IN ALL LOG MESSAGES FROM THIS FILE.
DOMAIN_NAME = "client_axis_vio"


class AxisVIOCoreClient(core_property_client.CorePropertyClient):
    # ===========================================
    # Abstract methods redefined here
    # ===========================================
    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(AXIS_VIO_SERVICE_NAME)

    # ===========================================
    # Static methods defined here
    # ===========================================
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return_val = node.type and node.type == AXIS_VIO_NODE_NAME
        return return_val

    def initialize(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initializing {AXIS_VIO_SERVICE_NAME} service")
        token = service.initialize(self.ctx, done_cb)
        return self.add_pending(token)

    # ===========================================
    # Core control
    # ===========================================

    def reset_core(self, done: DoneHWCommand = None):
        """
        Resets the VIO core and sets the probe out(s) to their initial value.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.
        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initiating AXIS VIO core reset")
        options = {"node_id": self.ctx}
        token = service.reset_core(options, done_cb)
        return self.add_pending(token)

    def get_core_info(self, done: DoneHWCommand = None):
        """
        Gets the core information properties associated with the VIO.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict:
                (Key, Value) - (Property name, Property value).
                Shown below are the keys that will be present in the return data along with
                example values

                ::

                    {
                     'core_type': 2, # int
                     'drv_ver': 0, # int
                     'tool_major_ver': 19, # int
                     'tool_minor_ver': 1, # int
                     'core_major_ver': 1, # int
                     'core_minor_ver': 0, # int
                     'uuid': "53E5F1C310B952D3BB10986C3375587A", # str
                    }

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initiating core info fetch")
        token = service.get_property_group(self.ctx, "core_info", done_cb)
        return self.add_pending(token)

    def get_ports_info(self, done: DoneHWCommand = None):
        """
        Gets static information like port counts, associated with the VIO core.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict:
                (Key, Value) - (Property name, Property value).
                Shown below are the keys that will be present in the return data along with
                example values

                ::

                    {
                        'port_in_count': 2, # int
                        'port_out_count': 3, # int

                         # Value at index 'X' is the width of port_in'X'
                         'port_in': [32, 64],

                         # Value at index 'X' is the width of port_out'X'
                         'port_out': [32, 64, 128]
                    }

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initiating ports info fetch")
        token = service.get_property_group(self.ctx, "static_info", done_cb)
        return self.add_pending(token)

    def get_control_signals(self, done: DoneHWCommand = None):
        """
        Gets the value of the control signals.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, bool):
                (Key, Value) - (Property name, Property value).
                Shown below are signals in the control group and example values for them.

                ::

                    {
                     'reset': False,
                     'commit': False,
                    }

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initiating control signal fetch")
        token = service.get_property_group(self.ctx, "control", done_cb)
        return self.add_pending(token)

    # ===========================================
    # Port related
    # ===========================================

    def commit_port_out_data(
        self, port_data: typing.Dict[int, typing.Union[str, int]], done: DoneHWCommand = None
    ):
        """
        Commit new value to port out(s).

        Args:
            port_data (dict): (Key, Value) - (Port out number, new data for the port).
                Default input format for port data is :class:`int`. However, data can be
                provided as **binary**, **hex** or **octal** using the prefixes
                ``0b``, ``0x`` and ``0o`` before the data.
                For example, to write a value of ``0x1122334455`` to 32-bit wide port out's,
                the arguments to the function can be

                ::

                    commit_port_out_data({
                                            0: 0x11223344,

                                            # Int equivalent of 0x11223344
                                            1: 287454020,

                                            # Binary equivalent of 0x11223344
                                            2: 0b10001001000100011001101000100,

                                            # Octal equivalent of 0x11223344
                                            3: 0o2110431504
                                        })

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            bool:
                ``True`` if data was written to the core, ``False`` otherwise

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initiating port out commit")
        options = {"port_data": port_data}
        token = service.commit_port_out_data(self.ctx, options, done_cb)
        return self.add_pending(token)

    def refresh_port_in_data(self, done: DoneHWCommand = None):
        """
        Refresh port in(s) and activity(if available) in the core object
        with value(s) from the core.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, list(int)):
                (Key, Value) - (Port type, Value for all available ports).
                For example, if the VIO design has 1 x Port in with activity,
                the return data would be

                ::

                    {
                     'In': [15],
                     'Activity': ['UDNB']
                    }

                     Interpreting activity
                     ---------------------
                     "B" signifies 1 to 0 + 0 to 1 transition for the bit
                     "R" signifies 0 to 1 transition for the bit
                     "F" signifies 1 to 0 transition for the bit
                     "N" signifies no change


                In this example, the value of ``Port in 0`` is the value at
                index ``0`` for key ``In``.

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initiating port in refresh")

        options = {"node_id": self.ctx}
        token = service.refresh_port_in_data(options, done_cb)
        return self.add_pending(token)

    def refresh_port_out_data(
        self, port_out_numbers: typing.Union[typing.List[int], int] = "", done: DoneHWCommand = None
    ):
        """
        Refresh port out(s) in the core object with value(s) from the core.

        Args:
            port_out_numbers (int, list(int)): The port out numbers

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, list(int)):
                (Key, Value) - (Port type, Value for all available ports).
                For example, if the VIO design has 2 x Port out's, the return data would be

                ::

                    # If command was vio.refresh_port_out_data()
                    # or vio.refresh_port_out_data([0, 1])
                    {
                     'Out': [15, 16],
                    }

                    # If command was vio.refresh_port_out_data(1)
                    {
                     'Out': [None, 16],
                    }

                    # If command was vio.refresh_port_out_data(0)
                    {
                     'Out': [15, None],
                    }

                     Interpreting activity
                     ---------------------
                     "B" signifies 1 to 0 + 0 to 1 transition for the bit
                     "R" signifies 0 to 1 transition for the bit
                     "F" signifies 1 to 0 transition for the bit
                     "N" signifies no change


                In this example, the value of ``Port out 0`` is the value at
                index ``0`` for key ``Out``.

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initiating port out commit")

        options = {"node_id": self.ctx, "port_out_numbers": port_out_numbers}
        token = service.refresh_port_out_data(options, done_cb)
        return self.add_pending(token)

    def get_port_data(self, done: DoneHWCommand = None):
        """
        Get current port value from the VIO properties.

        Args:
            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, list(int)):
                (Key, Value) - (Port type, Value for all available ports).
                For example, if the VIO design has 1 x Port in with activity and
                2 x Port out, the return data would be

                ::

                    {
                     'In': [15],
                     'Out': [287454020, 1432778632],
                     'Activity': ['UDNB']
                    }

                     Interpreting activity
                     ---------------------
                     "B" signifies 1 to 0 + 0 to 1 transition for the bit
                     "R" signifies 0 to 1 transition for the bit
                     "F" signifies 1 to 0 transition for the bit
                     "N" signifies no change


                In this example, the value of ``Port out 1`` is the value at
                index ``1`` for key ``Out``.

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info(f"Initiating port data fetch")
        options = {"node_id": self.ctx}
        token = service.get_port_data(options, done_cb)
        return self.add_pending(token)

    # ===========================================
    # Probe related
    # ===========================================

    def define_probe(
        self,
        probe_options: typing.Union[typing.Dict[str, str], typing.List[typing.Dict[str, str]]],
        done: DoneHWCommand = None,
    ):
        """
        Create a new probe.

        Args:
            probe_options (list(dict(str, str))):

                Its recommended to create a batch of probes. However, in case only one probe
                needs to be created, the probe_options can be provided as a ``dict(str, str]``.

                Each of the probe options need to have the following keys in the dictionary

                - ``name`` - Name of the new probe; This has to be **unique** for every probe
                - ``net`` - Name of the net the probe is mapped to
                - ``map`` - A white space separated list of port slices and constants in any order


                Following types of probes are currently supported

                    - Constant only
                    - Port in(s)
                    - Port in(s) + constant
                    - Port out(s)

                The below snippet, illustrates options supported by ``map``

                ::

                    "0101_1101"
                        Specifies a binary constant value. '_' may be used in between the constant
                        for readability purposes.

                    "port_out0[31:0]"
                        Specifies a 32-bit slice of a port. In the slice X:Y, X can be > or < Y.

                    "port_in0"
                        No explicit range means all bits of the physical port.
                        "port_in0" means same as "port_in0[31:0]", if the port is 32 bits wide.

                    "port_out0[4]"
                        Shorter notation for "port_out0[4:4]"

                    "0101 port_in0[31:0] 1101"
                        Specifies a constant and port slice. The constants need not be at the
                        beginning or end of the map.


            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info("Initiating probe creation")
        options = {"node_id": self.ctx, "probe_options": probe_options}
        token = service.define_probe(options, done_cb)
        return self.add_pending(token)

    def report_probe(self, probe_name: str = "", done: DoneHWCommand = None):
        """
        Report information for given probe name.

        Args:
            probe_name (str): (**Optional**) Name of the probe.
                If not provided, information about all available probe(s) is returned.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        Returns:
            dict(str, dict):
                (Key, Value) - (Probe name, probe information).
                The probe information dictionary has the following keys

                ::

                    {
                     'name': 'probeA',

                     'category': 'Out',

                     'net': 'dummy',

                     # Entry at index 'X', specifies the mapping for bit 'X' of the probe
                     'bits': ['port_out0_data[0]',
                              'port_out0_data[1]',
                              'port_out1_data[30]',
                              'port_out1_data[31]'],

                     # Bit length of the probe
                     'length': 4,

                     # If the probe has constants, this is True
                     'has_constants': False,

                     # The ports that have at least a bit mapped to the probe
                     'ports_part_of_probe': ['port_out0_data', 'port_out1_data'],

                     # Current value of probe as int
                     'value': 12,

                     # False if this probe is read only, True otherwise
                     'is_write_allowed': True
                    }

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info("Initiating probe info fetch")
        options = {"node_id": self.ctx, "probe_name": probe_name}
        token = service.report_probe(options, done_cb)
        return self.add_pending(token)

    def undefine_probe(
        self, probe_name: typing.Union[typing.List[str], str] = "All", done: DoneHWCommand = None
    ):
        """
        Remove a previously defined probe.

        Args:
            probe_name (list(str]): (**Optional**) Name of the probe as a :obj:`list` of :obj:`str`.
                 If removing a single probe, then the probe name can be provided as an :obj:`str`.
                 If not provided, all available probes are deleted.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info("Initiating probe removal destruction")
        options = {"node_id": self.ctx, "probe_name": probe_name}
        token = service.undefine_probe(options, done_cb)
        return self.add_pending(token)

    def refresh_probe(self, probe_name: str = "All", done: DoneHWCommand = None):
        """
        Refresh the value of the probe.

        Note:
            Refresh probe also refreshes the values of the ports.

        Args:
            probe_name (str): (**Optional**) Name of the probe.
                If not provided, all probes are refreshed.

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info("Initiating probe refresh")
        options = {"node_id": self.ctx, "probe_name": probe_name}
        token = service.refresh_probe(options, done_cb)
        return self.add_pending(token)

    def commit_probe(self, probe_options: typing.Dict[str, str], done: DoneHWCommand = None):
        """
        Commit new value to supported probe(s).

        Note:
            Currently commit is supported only for probe category of "Out" i.e.
            probe that has only port outs mapped to the probe bits.

        Args:
            probe_options (dict(str, str)): (Key, Value) - (Probe name, new value)
                Default input format for port data is :class:`int`. However, data can be
                provided as **binary**, **hex** or **octal** using the prefixes
                ``0b``, ``0x`` and ``0o`` before the data.

                For example, to write a value of ``0x1122334455`` to 32-bit wide probes,
                the arguments to the function can be

                ::

                    commit_probe({
                                    'probeA': 0x11223344,

                                    # Int equivalent of 0x11223344
                                    'probeB': 287454020,

                                    # Binary equivalent of 0x11223344
                                    'probeC': 0b10001001000100011001101000100,

                                    # Octal equivalent of 0x11223344
                                    'probeD': 0o2110431504,
                                 })

            done: **(Optional)** If callback is desired once operation is complete,
                then function/method should be provided.

        """
        service, done_cb = self.make_done(done)
        log[DOMAIN_NAME].info("Initiating probe addition")
        options = {"node_id": self.ctx, "probe_data": probe_options}
        token = service.commit_probe(options, done_cb)
        return self.add_pending(token)



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from chipscopy.tcf import protocol
from chipscopy.tcf.channel import ChannelListener
from chipscopy.proxies import DebugCoreProxy as dc_service
from chipscopy.utils.logger import log
from . import CsManager, Node, add_manager, get_manager, _managers, remove_manager
from .poll import DebugCorePollScheduler

MANAGER_TYPE = "debugcore"

csm_pattern = re.compile(r".+\." + MANAGER_TYPE)


def get_managers():
    return [csm for name, csm in _managers.items() if csm_pattern.match(name)]


def add_manager_from_channel(channel):
    serv = channel.getRemoteService(dc_service.DebugCoreService)
    if not serv:
        raise Exception("Server does not have DebugCore Service")

    # set up memory view manager
    name = "{}.{}".format(channel.remote_peer.getID(), MANAGER_TYPE)
    for manager in get_managers():
        if manager.name == name:
            if manager.channel != channel:
                remove_manager(manager)
                break
            return manager

    cs_manager = DebugCoreManager(name, channel)
    add_manager(cs_manager)

    # add channel listener for disconnection event
    channel.addChannelListener(ManagerChannelListener(cs_manager))

    # trigger update
    cs_manager.update_nodes()
    return cs_manager


def get_manager_from_channel(channel):
    channel_name = channel.remote_peer.getID()
    return get_manager(f"{channel_name}.{MANAGER_TYPE}")


def get_manager_from_channel_id(channel_id):
    return get_manager(f"{channel_id}.{MANAGER_TYPE}")


def id_domain_enable(param, param1):
    pass


class DebugNodeListener(dc_service.DebugNodeListener):
    def __init__(self, manager):
        self.manager = manager

    def node_added(self, node_ctx, props):
        if log.is_domain_enabled("dm", "DEBUG"):
            log.dm.debug(f"{self.manager.name}: Adding Node {node_ctx}: {props}")
        parent_ctx = props.get("ParentID")
        if not parent_ctx:
            parent_ctx = ""
        node = self.manager.add_node(node_ctx, parent_ctx)
        node.update(props)
        additional_props = props.get("additional_props")
        if additional_props and type(additional_props) == dict:
            node.update(additional_props)

    def node_changed(self, node_ctx, props):
        if id_domain_enable("dm", "DEBUG"):
            log.dm.debug(f"{self.manager.name}: Changing Node {node_ctx}: {props}")
        parent_ctx = props.get("ParentID")
        if not parent_ctx:
            parent_ctx = ""
        node = self.manager.add_node(node_ctx, parent_ctx)
        node.update(props)
        additional_props = props.get("additional_props")
        if additional_props and type(additional_props) == dict:
            node.update(additional_props)

    def node_removed(self, node_ctx):
        log.dm.debug(f"{self.manager.name}: Removing Node {node_ctx}")
        try:
            node = self.manager[node_ctx]
            # node.invalidate()
            self.manager.remove_node(node.ctx)
        except KeyError:
            pass


class DebugCoreManager(CsManager):
    def __init__(self, name: str = "", channel=None):
        super(DebugCoreManager, self).__init__(name, channel)
        self.dc = None
        self.poll_scheduler = None
        if self.channel:
            self.dc = self.channel.getRemoteService(dc_service.DebugCoreService)
            self.poll_scheduler = DebugCorePollScheduler(self.channel)
        if self.dc:
            self.dc.add_listener(DebugNodeListener(self))

    def update_nodes(self):
        self._update_children(self)
        self._update_polls()

    def _update_polls(self):
        def done_poll_update(token, error, results):
            self.remove_pending(token)

        if self.poll_scheduler:
            self.add_pending(self.poll_scheduler.update_polls(done=done_poll_update))

    def _update_children(self, parent):
        cs_manager = self

        def done_get_children(token, error, results):
            if not error and results:
                children = results
                if isinstance(children, str):
                    children = [children]
                cs_manager.set_children(parent, children)

                for ctx in children:
                    node = cs_manager.add_node(ctx, parent.ctx)
                    cs_manager._update_static_info(node)
                    cs_manager._update_children(node)
            cs_manager.remove_pending(token)

        if self.dc:
            self.add_pending(self.dc.get_children(parent.ctx, done_get_children))

    def _update_static_info(self, node: Node):
        def done_get_context(token, error, props):
            if not error:
                node.update(props)
            node.remove_pending(token)

        if self.dc:
            node.add_pending(self.dc.get_context(node.ctx, done_get_context))


class ChannelOpenListener(protocol.ChannelOpenListener):
    def onChannelOpen(self, channel):
        if channel.getRemoteService(dc_service.DebugCoreService):
            add_manager_from_channel(channel)


class ManagerChannelListener(ChannelListener):
    def __init__(self, cs_manager: DebugCoreManager):
        self.cs_manager = cs_manager

    def onChannelClosed(self, error):
        if self.cs_manager:
            self.cs_manager.invalidate()
            self.cs_manager.channel.removeChannelListener(self)


initialized = False


def init():
    global initialized
    if not initialized:
        protocol.invokeAndWait(protocol.addChannelOpenListener, ChannelOpenListener())
        initialized = True



# Copyright 2021 Xilinx, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union, Callable, Any
from chipscopy.tcf.services import DoneHWCommand, Service
from chipscopy.client.core_property_client import PropertyValues
from chipscopy.proxies.CorePropertyProxy import CorePropertyProxy

# Memory Models
MEMORY_MODEL_DDRMC = "memory_model_ddrmc"
MEMORY_MODEL_MILA_DDR = "memory_model_mila_ddr"

# DDRMC Memory Domains and Indexes
DDRMC_MAIN = "ddrmc_main"
DDRMC_NOC = "ddrmc_noc"
DDRMC_UB = "ddrmc_ub"

DDRMC_MAIN_INDEX = 0
DDRMC_NOC_INDEX = 1
DDRMC_UB_INDEX = 2
DDRMC_DOMAIN_COUNT = 3

# Service name.
NAME = "DDRMC"


class DDRMCService(Service):
    """TCF DDRMC Service interface."""

    def getName(self) -> str:
        """
        Get service name of this service.

        :returns: Service name :const:`NAME`.
        """
        return NAME


class DDRMCProxy(CorePropertyProxy, DDRMCService):
    def __init__(self, channel):
        super().__init__(channel)
        self._listeners = {}

    #
    #  General methods
    #
    def initialize(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("initialize", (node_id,), done)

    def terminate(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("terminate", (node_id,), done)

    def write32(
        self, node_id: str, write_addr: int, data: int, domain_index: int, done: DoneHWCommand
    ) -> None:
        """
        Write 32 bit register.

        :param node_id: Core context id.
        :param write_addr: Address.
        :param data: 32 bit data
        :param domain_index: 0 for DDRMC_MAIN, 1 for DDRMC_NOC
        """
        return self.send_command("write32", (node_id, write_addr, data, domain_index), done)

    def read32(
        self,
        node_id: str,
        read_addr: int,
        read_word_count: int,
        domain_index: int,
        done: DoneHWCommand,
    ) -> Union[List[int], int]:
        """
        Read 32 bit register

        :param node_id: Core context id.
        :param read_addr: address to read from.
        :param read_word_count: number of 32 bit words.
        :param domain_index: 0 for DDRMC_MAIN, 1 for DDRMC_NOC
        :return int value if read_word_count == 1. Otherwise list of int values.
        """
        return self.send_command(
            "read32", (node_id, read_addr, read_word_count, domain_index), done
        )

    #
    #  Service specific user interfaces
    #
    def refresh_property(
        self, node_id: str, prop_names: List[str], done: DoneHWCommand
    ) -> PropertyValues:
        return self.send_command("refreshProperty", (node_id, prop_names), done)

    def refresh_cal_status(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("refreshCalStatus", (node_id,), done)

    def refresh_cal_margin(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("refreshCalMargin", (node_id,), done)

    def refresh_health_status(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("refreshHealthStatus", (node_id,), done)

    def refresh_dqs_status(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("refreshDQSStatus", (node_id,), done)

    def disable_tracking(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("disableTracking", (node_id,), done)

    def enable_tracking(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("enableTracking", (node_id,), done)

    def get_cal_status(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("getCalStatus", (node_id,), done)

    def get_cal_margin_mode(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("getCalMarginMode", (node_id,), done)

    def get_margin_ps_factors(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("getMarginPsFactors", (node_id,), done)

    def is_clock_gated(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("isClockGated", (node_id,), done)

    def run_eye_scan(
        self,
        node_id: str,
        done: DoneHWCommand,
        progress: Callable[[Any, str or Exception, Any], None] = None,
    ) -> bool:
        return self.send_xicom_command("runEyeScan", (node_id,), done, progress)

    def get_eye_scan_data(self, node_id: str, done: DoneHWCommand) -> None:
        return self.send_command("getEyeScanData", (node_id,), done)



# *****************************************************************************
# * Copyright (c) 2011, 2013-2014, 2016 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *****************************************************************************

from .. import protocol, channel
from .task import Task


class DataCache(object):
    """
    Objects of this class are used to cache TCF remote data.
    The cache is asynchronous state machine. The states are:
     1. Valid - cache is in sync with remote data, use getError() and getData()
        to get cached data
     2. Invalid - cache is out of sync, start data retrieval by calling
        validate()
     3. Pending - cache is waiting result of a command that was sent to remote
        peer
     4. Disposed - cache is disposed and cannot be used to store data.

    A cache instance can be created on any data type that needs to be cached.
    Examples might be context children list, context properties, context state,
    memory data, register data, symbol, variable, etc.
    Clients of cache items can register for cache changes, but don't need to
    think about any particular events since that is handled by the cache item
    itself.

    A typical cache client should implement Runnable interface.
    The implementation of run() method should:

    Validate all cache items required for client task.
    If anything is invalid then client should not alter any shared data
    structures, should discard any intermediate results and register (wait) for
    changes of invalid cache instance(s) state.

    When cache item state changes, client is invoked again and full validation
    is restarted.
    Once everything is valid, client completes its task in a single dispatch
    cycle.

    Note: clients should never retain copies of remote data across dispatch
          cycles! Such data would get out of sync and compromise data
          consistency. All remote data and everything derived from remote data
          should be kept in cache items that implement proper event handling
          and can keep data consistent across dispatch cycles.
    """

    __error = None
    __valid = None
    __posted = False
    __disposed = False
    __data = None

    def __init__(self, channel):
        assert channel
        self._channel = channel
        self._command = None
        self.__waiting_list = None

    def post(self):
        if self.__posted:
            return
        if not self.__waiting_list:
            return
        protocol.invokeLater(self)
        self.__posted = True

    def isValid(self):
        """
        @return True if cache contains up-to-date data or error.
        """
        return self.__valid

    def isPending(self):
        """
        @return True if data retrieval command is in progress.
        """
        return self._command is not None

    def isDisposed(self):
        """
        @return True if cache is disposed.
        """
        return self.__disposed

    def getError(self):
        """
        @return error object if data retrieval ended with an error, or None if
        retrieval was successful.
        Note: It is prohibited to call this method when cache is not valid.
        """
        assert self.__valid
        return self.__error

    def getData(self):
        """
        @return cached data object.
        Note: It is prohibited to call this method when cache is not valid.
        """
        assert protocol.isDispatchThread()
        assert self.__valid
        return self.__data

    def __call__(self):
        """
        Notify waiting clients about cache state change and remove them from
        wait list.
        It is responsibility of clients to check if the state change was one
        they are waiting for.
        Clients are not intended to call this method.
        """
        assert protocol.isDispatchThread()
        self.__posted = False
        if self.__waiting_list:
            arr = self.__waiting_list
            self.__waiting_list = None
            for r in tuple(arr):
                if isinstance(r, DataCache):
                    r.post()
                elif isinstance(r, Task) and not r.isDone():
                    r.run()
                else:
                    r()

                arr.remove(r)

            if self.__waiting_list is None:
                self.__waiting_list = arr

    def wait(self, cb):
        """
        Add a client call-back to cache wait list.
        Client call-backs are activated when cache state changes.
        Call-backs are removed from waiting list after that.
        It is responsibility of clients to check if the state change was one
        they are waiting for.
        @param cb - a call-back object
        """
        assert protocol.isDispatchThread()
        assert not self.__disposed
        assert not self.__valid
        if cb and not self.isWaiting(cb):
            if self.__waiting_list is None:
                self.__waiting_list = []
            self.__waiting_list.append(cb)

    def isWaiting(self, cb):
        """
        Return True if a client call-back is waiting for state changes of this
        cache item.
        @param cb - a call-back object.
        @return True if 'cb' is in the wait list.
        """
        if not self.__waiting_list:
            return False
        return cb in self.__waiting_list

    def __validate(self):
        """
        Initiate data retrieval if the cache is not valid.
        @return True if the cache is already valid
        """
        assert protocol.isDispatchThread()
        if self.__disposed or self._channel.getState() != channel.STATE_OPEN:
            self._command = None
            self.__valid = True
            self.__error = None
            self.__data = None
        else:
            if self._command is not None:
                return False
            if not self.__valid and not self.startDataRetrieval():
                return False
        assert self.__valid
        assert self._command is None
        self.post()
        return True

    def validate(self, cb=None):
        """
        If the cache is not valid, initiate data retrieval and
        add a client call-back to cache wait list.
        Client call-backs are activated when cache state changes.
        Call-backs are removed from waiting list after that.
        It is responsibility of clients to check if the state change was one
        they are waiting for.
        If the cache is valid do nothing and return True.
        @param cb - a call-back object (optional)
        @return True if the cache is already valid
        """
        if not self.__validate():
            self.wait(cb)
            return False
        return True

    def start(self, command):
        """
        Start cache pending state.
        Pending state is when the cache is waiting for a TCF command to return
        results.
        @param command - TCF command handle.
        """
        assert not self.__valid
        assert command
        assert self._command is None
        self._command = command

    def done(self, command):
        """
        End cache pending state, but not mark the cache as valid.
        @param command - TCF command handle.
        """
        if self._command is not command:
            return
        assert not self.__valid
        self._command = None
        self.post()

    def set(self, token, error, data):
        """
        End cache pending state and mark the cache as valid.
        If 'token' != None, the data represent results from a completed
        command.
        The data should be ignored if current cache pending command is not same
        as 'token'.
        It can happen if the cache was reset or canceled during data retrieval.
        @param token - pending command handle or None.
        @param error - data retrieval error or None
        @param data - up-to-date data object
        """
        assert protocol.isDispatchThread()
        if token and self._command is not token:
            return
        self._command = None
        if not self.__disposed:
            assert not self.__valid
            if self._channel.getState() != channel.STATE_OPEN:
                self.__error = None
                self.__data = None
            self.__error = error
            self.__data = data
            self.__valid = True
        self.post()

    def reset(self, data=None):
        """
        Force cache to become valid, cancel pending data retrieval if data is
        provided.
        @param data - up-to-date data object (optional)
        """
        assert protocol.isDispatchThread()
        if data is not None and self._command is not None:
            self._command.cancel()
            self._command = None
        if not self.__disposed:
            self.__data = data
            self.__error = None
            self.__valid = data is not None
        self.post()

    def cancel(self):
        """
        Invalidate the cache.
        Cancel pending data retrieval if any.
        """
        self.reset()
        if self._command is not None:
            self._command.cancel()
            self._command = None

    def dispose(self):
        """
        Dispose the cache.
        Cancel pending data retrieval if any.
        """
        self.cancel()
        self.__valid = True
        self.__disposed = True

    def __str__(self):
        res = '['
        if self.__valid:
            res += 'valid,'
        if self.__disposed:
            res += 'disposed,'
        if self.__posted:
            res += 'posted,'
        if self.__error is not None:
            res += 'error,'
        res += 'data='
        res += str(self.__data)
        res += ']'
        return res

    def startDataRetrieval(self):
        """
        Sub-classes should override this method to implement actual data
        retrieval logic.
        @return True is all done, False if retrieval is in progress.
        """
        raise NotImplementedError("Abstract method")
